from pathlib import Path
import importlib
import sys

import pandas as pd
import pytest
from click.testing import CliRunner

from abel_edge.config import load_config
from abel_edge.cli import main
from abel_edge.engine.ledger import read_trade_log, write_trade_log
from abel_edge.engine.trader import paper_run_one


ENGINE_CODE = """
from __future__ import annotations

import numpy as np
import pandas as pd
from datetime import date, datetime, timezone

from abel_edge.engine.base import StrategyEngine


class PaperDemoEngine(StrategyEngine):
    def compute_signals(self):
        bars = self.load_bars(limit=10)
        target = bars[bars['symbol'] == self.context.get('asset', 'ETHUSD')].copy().sort_values('timestamp')
        prices = target['close'].astype(float).to_numpy()
        dates = pd.DatetimeIndex(target['timestamp'])
        prev_close = pd.Series(prices).shift(1)
        positions = (prev_close >= 100.0).astype(float).fillna(0.0).to_numpy()
        return positions, dates, prices

    def get_latest_signal(self):
        positions, dates, prices = self.compute_signals()
        return {'position': float(positions[-1]), 'date': str(dates[-1]), 'price': float(prices[-1])}

    def get_paper_signal(self, *, as_of=None):
        bars = self.load_bars(limit=10, end=as_of)
        target = bars[bars['symbol'] == self.context.get('asset', 'ETHUSD')].copy().sort_values('timestamp')
        last_close = float(target['close'].iloc[-1])
        return {
            'next_position': 1.0 if last_close >= 100.0 else 0.0,
            'date': str(target['timestamp'].iloc[-1]),
            'price': last_close,
            'data_backend': 'csv',
            'data_fetch_status': 'provider_fetch',
            'data_latest_timestamp': str(target['timestamp'].iloc[-1]),
            'broker_decision_time': datetime(2026, 1, 4, 16, 30, tzinfo=timezone.utc),
            'provider_session_date': date(2026, 1, 4),
            'provider_sequence': np.int64(len(target)),
            'strategy_label': self.context.get('id'),
            'paper_audit_status': 'provider_fetch',
            'source': 'bad-source-override',
            'pnl': 999.0,
            'cum_return': 999.0,
            'gross_pnl': 999.0,
            'turnover': 999.0,
            'execution_cost': 999.0,
        }
""".strip()


def _write_strategy_project(tmp_path: Path) -> None:
    for name in list(sys.modules):
        if name == "strategies" or name.startswith("strategies."):
            sys.modules.pop(name, None)
    importlib.invalidate_caches()
    (tmp_path / "strategies").mkdir()
    (tmp_path / "strategies" / "__init__.py").write_text("", encoding="utf-8")
    (tmp_path / "strategies" / "paper_demo").mkdir()
    (tmp_path / "strategies" / "paper_demo" / "__init__.py").write_text("", encoding="utf-8")
    (tmp_path / "strategies" / "paper_demo" / "engine.py").write_text(
        ENGINE_CODE, encoding="utf-8"
    )
    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "ethusd.csv").write_text(
        "timestamp,close\n"
        "2026-01-01T00:00:00Z,100\n"
        "2026-01-02T00:00:00Z,110\n"
        "2026-01-03T00:00:00Z,90\n"
        "2026-01-04T00:00:00Z,120\n",
        encoding="utf-8",
    )
    (tmp_path / "data" / "btcusd.csv").write_text(
        "timestamp,close\n"
        "2026-01-01T00:00:00Z,50\n"
        "2026-01-02T00:00:00Z,70\n"
        "2026-01-03T00:00:00Z,130\n"
        "2026-01-04T00:00:00Z,80\n",
        encoding="utf-8",
    )
    (tmp_path / "strategies.yaml").write_text(
        """
settings:
  price_data:
    default_source: csv
    default_timeframe: 1d
strategies:
  - id: paper_demo
    name: "Paper Demo"
    asset: ETHUSD
    color: "#2563EB"
    engine: strategies.paper_demo.engine
    trade_log: data/trade_log_paper_demo.csv
    paper_log: data/paper_log_paper_demo.csv
    price_data:
      source: csv
      path: data/ethusd.csv
  - id: paper_alt
    name: "Paper Alt"
    asset: BTCUSD
    color: "#F59E0B"
    engine: strategies.paper_demo.engine
    trade_log: data/trade_log_paper_alt.csv
    paper_log: data/paper_log_paper_alt.csv
    price_data:
      source: csv
      path: data/btcusd.csv
""".strip()
        + "\n",
        encoding="utf-8",
    )


def _write_bootstrap_log(path: str, *, closes: list[float], positions: list[float]) -> None:
    dates = pd.to_datetime(["2026-01-01T00:00:00Z", "2026-01-02T00:00:00Z"], utc=True)
    returns = [0.0, closes[1] / closes[0] - 1.0]
    pnl = [position * value for position, value in zip(positions, returns)]
    write_trade_log(
        dates,
        returns,
        pnl,
        positions,
        path,
        close_prices=closes,
    )


def test_paper_appends_live_rows_with_close_fill_semantics(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root = Path.cwd()
        _write_strategy_project(root)
        sys.path.insert(0, str(root))

        dates = pd.to_datetime(["2026-01-01T00:00:00Z", "2026-01-02T00:00:00Z"], utc=True)
        returns = [0.0, 0.10]
        pnl = [0.0, 0.10]
        positions = [0.0, 1.0]
        write_trade_log(
            dates,
            returns,
            pnl,
            positions,
            "data/trade_log_paper_demo.csv",
            close_prices=[100.0, 110.0],
        )

        result = runner.invoke(
            main,
            ["paper", "--strategy", "paper_demo", "--as-of", "2026-01-04T00:00:00Z"],
        )

        assert result.exit_code == 0, result.output
        trade_df = read_trade_log("data/trade_log_paper_demo.csv")
        paper_df = read_trade_log("data/paper_log_paper_demo.csv")
        assert len(trade_df) == 2
        assert list(paper_df["source"].tail(2)) == ["live", "live"]
        assert list(paper_df["date"].dt.strftime("%Y-%m-%d").tail(2)) == [
            "2026-01-03",
            "2026-01-04",
        ]
        assert list(paper_df["position"].tail(2).round(2)) == [1.0, 0.0]
        assert list(paper_df["next_position"].tail(2).round(2)) == [0.0, 1.0]
        assert list(paper_df["close"].tail(2).round(2)) == [90.0, 120.0]
        assert float(paper_df.iloc[-2]["pnl"]) < 0
        assert paper_df.iloc[-1]["data_backend"] == "csv"
        assert paper_df.iloc[-1]["data_fetch_status"] == "provider_fetch"
        assert str(paper_df.iloc[-1]["data_latest_timestamp"]).startswith("2026-01-04")
        assert str(paper_df.iloc[-1]["broker_decision_time"]).startswith("2026-01-04T16:30:00")
        assert str(paper_df.iloc[-1]["provider_session_date"]) == "2026-01-04"
        assert int(paper_df.iloc[-1]["provider_sequence"]) == 4
        assert paper_df.iloc[-1]["paper_audit_status"] == "provider_fetch"
        assert paper_df.iloc[-1]["strategy_label"] == "paper_demo"
        assert paper_df.iloc[-1]["source"] == "live"
        assert float(paper_df.iloc[-1]["pnl"]) == 0.0
        assert float(paper_df.iloc[-1]["cum_return"]) != 999.0
        for column in ["gross_pnl", "turnover", "execution_cost"]:
            if column in paper_df.columns:
                assert float(paper_df.iloc[-1][column]) != 999.0
        sys.path.pop(0)


def test_paper_run_one_matches_golden_append_math(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root = Path.cwd()
        _write_strategy_project(root)
        sys.path.insert(0, str(root))

        try:
            _write_bootstrap_log(
                "data/trade_log_paper_demo.csv",
                closes=[100.0, 110.0],
                positions=[0.0, 1.0],
            )
            cfg = load_config()
            strategy = next(item for item in cfg["strategies"] if item["id"] == "paper_demo")

            result = paper_run_one(
                strategy,
                settings=cfg.get("settings"),
                as_of="2026-01-04T00:00:00Z",
            )

            paper_df = read_trade_log("data/paper_log_paper_demo.csv")
            live = paper_df[paper_df["source"] == "live"].reset_index(drop=True)
            assert result["n_rows"] == 2
            assert list(live["date"].dt.strftime("%Y-%m-%d")) == ["2026-01-03", "2026-01-04"]
            assert list(live["close"]) == pytest.approx([90.0, 120.0])
            assert list(live["position"]) == pytest.approx([1.0, 0.0])
            assert list(live["next_position"]) == pytest.approx([0.0, 1.0])
            assert list(live["asset_return"]) == pytest.approx([
                90.0 / 110.0 - 1.0,
                120.0 / 90.0 - 1.0,
            ])
            assert list(live["pnl"]) == pytest.approx([
                1.0 * (90.0 / 110.0 - 1.0),
                0.0 * (120.0 / 90.0 - 1.0),
            ])
            assert result["latest_snapshot"]["last_processed_date"] == "2026-01-04T00:00:00+00:00"
            assert result["latest_snapshot"]["current_position"] == 0.0
            assert result["latest_snapshot"]["next_position"] == 1.0
            assert result["latest_snapshot"]["latest_close"] == 120.0
        finally:
            sys.path.pop(0)


def test_paper_run_one_returns_latest_snapshot_for_each_strategy(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root = Path.cwd()
        _write_strategy_project(root)
        sys.path.insert(0, str(root))

        try:
            _write_bootstrap_log(
                "data/trade_log_paper_demo.csv",
                closes=[100.0, 110.0],
                positions=[0.0, 1.0],
            )
            _write_bootstrap_log(
                "data/trade_log_paper_alt.csv",
                closes=[50.0, 70.0],
                positions=[0.0, 0.0],
            )
            cfg = load_config()
            strategies = {item["id"]: item for item in cfg["strategies"]}

            demo = paper_run_one(
                strategies["paper_demo"],
                settings=cfg.get("settings"),
                as_of="2026-01-04T00:00:00Z",
            )
            alt = paper_run_one(
                strategies["paper_alt"],
                settings=cfg.get("settings"),
                as_of="2026-01-04T00:00:00Z",
            )

            expected_demo = {
                "strategy_id": "paper_demo",
                "asset": "ETHUSD",
                "last_processed_date": "2026-01-04T00:00:00+00:00",
                "current_position": 0.0,
                "next_position": 1.0,
                "latest_close": 120.0,
                "source": "live",
                "strategy_label": "paper_demo",
                "data_backend": "csv",
                "data_fetch_status": "provider_fetch",
                "data_latest_timestamp": "2026-01-04 00:00:00+00:00",
                "paper_audit_status": "provider_fetch",
            }
            for key, value in expected_demo.items():
                assert demo["latest_snapshot"][key] == value
            assert int(demo["latest_snapshot"]["provider_sequence"]) == 4

            expected_alt = {
                "strategy_id": "paper_alt",
                "asset": "BTCUSD",
                "last_processed_date": "2026-01-04T00:00:00+00:00",
                "current_position": 1.0,
                "next_position": 0.0,
                "latest_close": 80.0,
                "source": "live",
                "strategy_label": "paper_alt",
                "data_backend": "csv",
                "data_fetch_status": "provider_fetch",
                "data_latest_timestamp": "2026-01-04 00:00:00+00:00",
                "paper_audit_status": "provider_fetch",
            }
            for key, value in expected_alt.items():
                assert alt["latest_snapshot"][key] == value
            assert int(alt["latest_snapshot"]["provider_sequence"]) == 4
        finally:
            sys.path.pop(0)


def test_paper_is_idempotent_when_no_new_bars(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root = Path.cwd()
        _write_strategy_project(root)
        sys.path.insert(0, str(root))

        dates = pd.to_datetime(["2026-01-01T00:00:00Z", "2026-01-02T00:00:00Z"], utc=True)
        returns = [0.0, 0.10]
        pnl = [0.0, 0.10]
        positions = [0.0, 1.0]
        write_trade_log(
            dates,
            returns,
            pnl,
            positions,
            "data/trade_log_paper_demo.csv",
            close_prices=[100.0, 110.0],
        )

        first = runner.invoke(
            main,
            ["paper", "--strategy", "paper_demo", "--as-of", "2026-01-04T00:00:00Z"],
        )
        assert first.exit_code == 0, first.output

        second = runner.invoke(
            main,
            ["paper", "--strategy", "paper_demo", "--as-of", "2026-01-04T00:00:00Z"],
        )
        assert second.exit_code == 0, second.output
        assert "no new closed bars" in second.output

        trade_df = read_trade_log("data/trade_log_paper_demo.csv")
        paper_df = read_trade_log("data/paper_log_paper_demo.csv")
        assert len(trade_df) == 2
        assert len(paper_df) == 2
        assert int((paper_df["source"] == "live").sum()) == 2
        sys.path.pop(0)



"""Raw evaluation helpers for strategy experiments."""

from abel_edge.research.evaluate import (
    check_look_ahead,
    compute_k,
    render_validation_markdown,
    run_evaluation,
    write_evaluation_outputs,
)
from abel_edge.research.artifact_export import (
    export_strategy_artifact_zip,
    write_backtest_trade_log_from_metric_input,
)
from abel_edge.research.handoff import (
    HANDOFF_CONTRACT,
    build_strategy_handoff,
    load_strategy_handoff,
    validate_strategy_handoff,
)

__all__ = [
    "check_look_ahead",
    "compute_k",
    "render_validation_markdown",
    "run_evaluation",
    "write_evaluation_outputs",
    "export_strategy_artifact_zip",
    "write_backtest_trade_log_from_metric_input",
    "HANDOFF_CONTRACT",
    "build_strategy_handoff",
    "load_strategy_handoff",
    "validate_strategy_handoff",
]
from abel_edge.research.state_intent import (
    STATE_INTENT_SCHEMA,
    StateIntentEntry,
    validate_state_intent,
    validate_state_intent_path,
)

__all__ = [
    "STATE_INTENT_SCHEMA",
    "StateIntentEntry",
    "validate_state_intent",
    "validate_state_intent_path",
]



"""Command-line interface for abel-edge."""

from __future__ import annotations

# Hardening must run before numpy/joblib import (prevents fork/threads
# deadlock — cron 2.4h 2026-04-17, CLI 5h 2026-04-18). E402 is intentional.
from abel_edge._runtime_harden import apply as _apply_runtime_harden

_apply_runtime_harden()
import json  # noqa: E402
from pathlib import Path  # noqa: E402

import click  # noqa: E402

from abel_edge import __version__  # noqa: E402
from abel_edge.cache_cli import warm_cache as warm_cache_command  # noqa: E402
from abel_edge.cli_support import build_bars_loader  # noqa: E402
from abel_edge.research.cli import debug_evaluate as debug_evaluate_command  # noqa: E402
from abel_edge.research.cli import evaluate as evaluate_command  # noqa: E402
from abel_edge.research.cli import export_artifact as export_artifact_command  # noqa: E402
from abel_edge.research.cli import probe_data as probe_data_command  # noqa: E402
from abel_edge.research.cli import verify_data as verify_data_command  # noqa: E402
from abel_edge.research.cli import validate_handoff as validate_handoff_command  # noqa: E402

CONFIG_OPTION_HELP = (
    "Config file path (defaults to strategies.local.yaml if present, else strategies.yaml)"
)


def _get_version() -> str:
    """Return the framework version from the package source."""
    return __version__


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(version=_get_version(), prog_name="abel-edge")
def main():
    """abel-edge: Agent-native quant framework."""
    # Tree-kill trap + opt-in wall-clock timeout from env.
    # No-op for help/version (click short-circuits before this runs).
    from abel_edge._runtime_harden import install_from_env

    install_from_env()


main.add_command(evaluate_command)
main.add_command(export_artifact_command)
main.add_command(debug_evaluate_command)
main.add_command(probe_data_command)
main.add_command(verify_data_command)
main.add_command(validate_handoff_command)
main.add_command(warm_cache_command)


@main.command("version")
def version():
    """Show abel-edge version."""
    click.echo(f"abel-edge, version {_get_version()}")


@main.command()
@click.argument("name")
def init(name):
    """Scaffold a new abel-edge project."""
    from abel_edge.scaffold import scaffold_project

    try:
        root = scaffold_project(name)
    except FileExistsError as e:
        raise click.ClickException(str(e))

    click.echo(f"Created {root}/")
    click.echo("  strategies.yaml          (3 local sample-data demo strategies)")
    click.echo("  strategies/sma_crossover (DecisionContext SMA demo)")
    click.echo("  strategies/momentum_ml   (DecisionContext walk-forward GBDT demo)")
    click.echo("  strategies/feed_overlay_demo (DecisionContext auxiliary-feed demo)")
    click.echo("  CLAUDE.md + AGENTS.md    (agent harness)")
    click.echo("  .env.example             (Abel API key, optional)")
    click.echo("")
    click.echo("Note:")
    click.echo("  This is a standalone abel-edge project scaffold with local sample data.")
    click.echo("  For branch research inside an Abel-alpha workspace, use the Abel-alpha branch flow.")
    click.echo("")
    click.echo("Next:")
    click.echo(f"  cd {name}")
    click.echo("  abel-edge run")
    click.echo("  abel-edge dashboard")
    click.echo("  abel-edge validate")


@main.command()
@click.option(
    "--env-path", default=".env", show_default=True, help="File path for storing ABEL_API_KEY"
)
@click.option(
    "--no-browser", is_flag=True, help="Print the authorization URL without opening a browser"
)
@click.option(
    "--json",
    "json_output",
    is_flag=True,
    help="Emit newline-delimited JSON events and the final result",
)
@click.option("--print-token", is_flag=True, help="Include the resolved API key in command output")
@click.option("--force", is_flag=True, help="Ignore any existing ABEL_API_KEY and run OAuth again")
@click.option(
    "--timeout",
    default=300,
    show_default=True,
    type=click.IntRange(min=1),
    help="Authorization timeout in seconds",
)
def login(env_path, no_browser, json_output, print_token, force, timeout):
    """Run explicit Abel OAuth and optionally persist the API key to .env."""
    try:
        from abel_edge.plugins.abel.auth import login_with_oauth
    except ImportError:
        raise click.ClickException("Abel plugin not installed. See: abel_edge/plugins/AGENTS.md")

    def _notify(message: str) -> None:
        click.echo(message, err=True)
        click.get_text_stream("stderr").flush()

    def _emit_json_event(payload: dict[str, object]) -> None:
        stream = click.get_text_stream("stdout")
        click.echo(json.dumps(payload, sort_keys=True), file=stream)
        stream.flush()

    try:
        result = login_with_oauth(
            env_path=env_path,
            open_browser=not no_browser,
            timeout_seconds=timeout,
            force=force,
            notify=_notify,
            on_handoff=_emit_json_event if json_output else None,
            on_pending=_emit_json_event if json_output else None,
        )
    except Exception as e:
        raise click.ClickException(str(e))

    output = dict(result)
    if not print_token:
        output.pop("api_key", None)

    if json_output:
        _emit_json_event(output)
        return

    if result["status"] == "already_configured":
        source = str(result.get("source") or "").strip()
        source_path = str(result.get("source_path") or "").strip()
        if source == "shared_auth_file" and source_path:
            click.echo(f"Reusing existing Abel auth from shared file: {source_path}")
        elif source == "project_env" and source_path:
            click.echo(f"Reusing existing Abel auth from project env: {source_path}")
        elif source == "env_var":
            click.echo("Reusing existing Abel auth from the current process environment.")
        else:
            click.echo("Abel API key already configured.")
        if print_token:
            click.echo(output["api_key"])
        return

    click.echo(f"Abel API key saved to {env_path}.")
    if print_token:
        click.echo(output["api_key"])


@main.command()
@click.option("--strategy", default=None, help="Run a specific strategy by ID")
@click.option("--config", default=None, help=CONFIG_OPTION_HELP)
def run(strategy, config):
    """Run strategies and write trade logs."""
    from abel_edge.config import load_config
    from abel_edge.engine.trader import run_all

    cfg = load_config(config)
    if not cfg["strategies"]:
        click.echo("No strategies configured. Add strategies to strategies.yaml.")
        return

    click.echo(f"Running {len(cfg['strategies'])} strategies...")
    results = run_all(cfg, strategy_id=strategy)
    click.echo(f"Done. {len(results)} strategies executed.")


@main.command("paper")
@click.option("--strategy", default=None, help="Paper-trade a specific strategy by ID")
@click.option("--config", default=None, help=CONFIG_OPTION_HELP)
@click.option("--as-of", default=None, help="Only process bars up to this timestamp")
def paper(strategy, config, as_of):
    """Append live paper-trading rows using the latest closed bars."""
    from abel_edge.config import load_config
    from abel_edge.engine.trader import paper_run_all

    cfg = load_config(config)
    if not cfg["strategies"]:
        click.echo("No strategies configured. Add strategies to strategies.yaml.")
        return

    click.echo(f"Paper trading {len(cfg['strategies'])} strategies...")
    results = paper_run_all(cfg, strategy_id=strategy, as_of=as_of)
    click.echo(f"Done. {len(results)} strategies processed.")


@main.command()
@click.option("--config", default=None, help=CONFIG_OPTION_HELP)
@click.option("--output", default="dashboard.html", help="Output HTML path")
def dashboard(config, output):
    """Generate dashboard HTML."""
    from abel_edge.dashboard.generator import generate
    from abel_edge.config import load_config

    cfg = load_config(config)
    bars_loader = build_bars_loader(cfg)
    generate(config, output, bars_loader=bars_loader)
    click.echo(f"Dashboard generated: {output}")


@main.command("signal-demo")
@click.option("--config", default=None, help=CONFIG_OPTION_HELP)
@click.option("--strategy", required=True, help="Render a specific strategy signal demo page")
@click.option("--output", default="signal-demo.html", help="Output HTML path")
def signal_demo(config, strategy, output):
    """Generate a single-strategy Signal Demo page."""
    from abel_edge.dashboard.generator import generate_signal_demo
    from abel_edge.config import load_config

    cfg = load_config(config)
    bars_loader = build_bars_loader(cfg)

    try:
        generate_signal_demo(config, output, strategy_id=strategy, bars_loader=bars_loader)
    except ValueError as e:
        raise click.ClickException(str(e))
    click.echo(f"Signal demo generated: {output}")


@main.command("tracking")
@click.option("--config", default=None, help=CONFIG_OPTION_HELP)
@click.option("--strategy", required=True, help="Render a specific strategy tracking page")
@click.option("--output", default="tracking.html", help="Output HTML path")
def tracking(config, strategy, output):
    """Generate a single-strategy dashboard focused on paper tracking."""
    from abel_edge.dashboard.generator import generate_tracking_page
    from abel_edge.config import load_config

    cfg = load_config(config)
    bars_loader = build_bars_loader(cfg)

    try:
        generate_tracking_page(config, output, strategy_id=strategy, bars_loader=bars_loader)
    except ValueError as e:
        raise click.ClickException(str(e))
    click.echo(f"Tracking page generated: {output}")


@main.command()
@click.option("--strategy", default=None, help="Validate a specific strategy by ID")
@click.option("--verbose", is_flag=True, help="Show detailed failure info")
@click.option(
    "--csv", "csv_path", default=None, help="Validate a standalone CSV (date,pnl columns)"
)
@click.option(
    "--dsr-trials",
    type=click.IntRange(min=1),
    default=None,
    help="Declared strategy exploration count used by DSR (overrides profile default)",
)
@click.option("--export", "export_path", default=None, help="Export report to file")
@click.option("--config", default=None, help=CONFIG_OPTION_HELP)
def validate(strategy, verbose, csv_path, dsr_trials, export_path, config):
    """Run Abel Proof validation on strategies."""
    import io
    import sys

    from abel_edge.validation.gate import validate_strategy, print_validation_report

    results = {}
    old_stdout = None
    capture = None

    if csv_path:
        # Quick path: validate a standalone CSV without strategies.yaml
        if not Path(csv_path).exists():
            raise click.ClickException(f"CSV not found: {csv_path}")
        result = validate_strategy(csv_path, dsr_trials=dsr_trials)
        results[Path(csv_path).stem] = result
    else:
        from abel_edge.config import load_config

        cfg = load_config(config)
        strategies_list = cfg["strategies"]
        if strategy:
            strategies_list = [s for s in strategies_list if s["id"] == strategy]

        if not strategies_list:
            click.echo("No strategies to validate.")
            return

        for s_cfg in strategies_list:
            sid = s_cfg["id"]
            log_path = s_cfg.get("trade_log", "")
            if not Path(log_path).exists():
                results[sid] = {
                    "verdict": "SKIP",
                    "score": "0/0",
                    "failures": [f"Trade log not found: {log_path}. Run 'abel-edge run' first."],
                    "metrics": {},
                    "triangle": {"ratio": 0, "rank": 0, "shape": 0},
                    "profile": "unknown",
                }
                continue
            results[sid] = validate_strategy(log_path, dsr_trials=dsr_trials)

    # Capture output for --export
    if export_path:
        old_stdout = sys.stdout
        sys.stdout = capture = io.StringIO()

    try:
        print_validation_report(results)

        if verbose:
            print()
            for sid, r in results.items():
                if r.get("metrics"):
                    print(f"  {sid} metrics:")
                    m = r["metrics"]
                    for key in (
                        "sharpe",
                        "lo_adjusted",
                        "sortino",
                        "total_return",
                        "annual_return",
                        "max_dd",
                        "calmar",
                        "dsr",
                        "dsr_trials_used",
                        "omega",
                        "position_ic",
                        "position_hit_rate",
                    ):
                        if key.startswith("position_") and not m.get(
                            "position_ic_applicable", False
                        ):
                            continue
                        if key in m:
                            print(f"    {key:20s} {m[key]:.4f}")
                    if "yearly_sharpes" in m:
                        print("    yearly_sharpes:")
                        for yr, sh in sorted(m["yearly_sharpes"].items()):
                            print(f"      {yr}: {sh:.2f}")
    finally:
        if export_path:
            sys.stdout = old_stdout
            report_text = capture.getvalue() if capture is not None else ""
            click.echo(report_text, nl=False)  # also print to terminal
            Path(export_path).write_text(report_text, encoding="utf-8")
            click.echo(f"\n  Report exported to {export_path}")

    all_pass = all(r["verdict"] in ("PASS", "SKIP") for r in results.values())
    sys.exit(0 if all_pass else 1)


@main.command()
@click.argument("ticker")
@click.option(
    "--mode",
    type=click.Choice(["parents", "mb", "all"]),
    default="parents",
    show_default=True,
)
@click.option(
    "--limit",
    default=10,
    show_default=True,
    type=int,
    help="Maximum nodes to return (hard cap 20)",
)
@click.option("--json", "json_output", is_flag=True, help="Emit structured JSON output")
def discover(ticker, mode, limit, json_output):
    """Discover causal graph nodes for an asset via Abel API."""
    try:
        from abel_edge.plugins.abel.discover import (
            discover_graph_nodes,
            discover_graph_payload,
        )
    except ImportError:
        raise click.ClickException("Abel plugin not installed. See: abel_edge/plugins/AGENTS.md")
    try:
        if json_output:
            output = discover_graph_payload(ticker, mode=mode, limit=limit)
            click.echo(json.dumps(output, indent=2, sort_keys=True))
            return
        output = discover_graph_nodes(ticker, mode=mode, limit=limit)
    except Exception as e:
        raise click.ClickException(str(e))
    click.echo(output)


@main.command()
@click.option("--config", default=None, help=CONFIG_OPTION_HELP)
def status(config):
    """Show strategy status summary."""
    from abel_edge.config import load_config

    cfg = load_config(config)
    click.echo(f"Strategies: {len(cfg['strategies'])}")
    for s in cfg["strategies"]:
        click.echo(f"  {s['name']:20s}  {s['asset']:6s}  {s.get('badge', '?')}")


if __name__ == "__main__":
    main()



"""ETHUSD causal strategy using Abel discovery output."""

from __future__ import annotations

import json
from pathlib import Path

import numpy as np
import pandas as pd

from abel_edge.engine.base import StrategyEngine

GRAPH_PATH = Path(__file__).parent / "causal_graph.json"
CONVICTION_MIN = 0.75
DEFAULT_LAG = 1
DEFAULT_WINDOW = 5


class ETHUSDCausalEngine(StrategyEngine):
    """Vote-based causal strategy for ETHUSD."""

    def __init__(self, context: dict | None = None, n_days: int = 600) -> None:
        super().__init__(context=context)
        self.n_days = n_days
        with open(GRAPH_PATH, encoding="utf-8") as f:
            self.graph = json.load(f)

    def compute_signals(self):
        positions, dates, target_prices, _ = self._compute_history()
        return positions, dates, target_prices

    def get_paper_signal(self, *, as_of=None):
        positions, dates, target_prices, components = self._compute_history(as_of=as_of)
        if len(dates) == 0:
            raise ValueError("No bars available to compute paper signal.")
        next_position = self._compute_next_position(components, pd.DatetimeIndex(dates))
        return {
            "next_position": float(next_position),
            "date": str(dates[-1]),
            "price": float(target_prices[-1]),
        }

    def _compute_history(self, as_of=None):
        components = [self._normalize_component(item) for item in self.graph.get("parents", [])]
        symbols = [self.context.get("asset", "ETHUSD")] + [comp["ticker"] for comp in components]
        bars = self.load_bars(symbols=symbols, limit=self.n_days, end=as_of)

        target_symbol = symbols[0]
        target = bars[bars["symbol"] == target_symbol].copy()
        if len(target) < 30:
            raise ValueError(f"Not enough price data for {target_symbol}: need 30+ bars.")

        target = target.sort_values("timestamp").tail(self.n_days)
        target_prices = target["close"].astype(float).to_numpy()
        dates = pd.DatetimeIndex(target["timestamp"])

        sig_matrix = []
        aligned_returns = []
        for comp in components:
            tau = comp["lag"]
            win = comp["window"]
            comp_bars = bars[bars["symbol"] == comp["ticker"]].copy()
            comp_bars = comp_bars.sort_values("timestamp")
            aligned = comp_bars.set_index("timestamp")["close"].reindex(dates).ffill()
            ret = aligned.pct_change().fillna(0.0)
            aligned_returns.append(ret)
            if win > 1:
                sig = np.sign(ret.rolling(win).sum().shift(tau)).values
            else:
                sig = np.sign(ret.shift(tau)).values
            sig_matrix.append(np.nan_to_num(sig, nan=0.0))

        sig_matrix = np.array(sig_matrix)
        n_days = len(dates)
        n_up = (sig_matrix > 0).sum(axis=0)
        n_down = (sig_matrix < 0).sum(axis=0)
        n_active = (sig_matrix != 0).sum(axis=0)
        vote_frac = np.divide(
            n_up,
            n_active,
            out=np.full(n_days, 0.5, dtype=float),
            where=n_active > 0,
        )

        positions = np.zeros(n_days)
        bull = n_up > n_down
        positions[bull] = vote_frac[bull] ** 2
        positions[bull & (vote_frac < CONVICTION_MIN)] = 0.0
        return (
            np.maximum(positions, 0.0),
            dates,
            target_prices,
            list(zip(components, aligned_returns)),
        )

    def get_latest_signal(self):
        positions, dates, prices = self.compute_signals()
        return {
            "position": float(positions[-1]),
            "date": str(dates[-1].date()),
            "price": float(prices[-1]),
        }

    def _compute_next_position(self, component_returns, dates: pd.DatetimeIndex) -> float:
        if len(dates) == 0 or not component_returns:
            return 0.0

        signals = []
        for comp, ret in component_returns:
            signal = self._next_component_signal(ret, lag=comp["lag"], window=comp["window"])
            signals.append(signal)

        sig_values = np.asarray(signals, dtype=float)
        n_up = int(np.sum(sig_values > 0))
        n_down = int(np.sum(sig_values < 0))
        n_active = int(np.sum(sig_values != 0))
        if n_up <= n_down or n_active == 0:
            return 0.0

        vote_frac = n_up / n_active
        if vote_frac < CONVICTION_MIN:
            return 0.0
        return max(vote_frac**2, 0.0)

    def _next_component_signal(self, ret: pd.Series, *, lag: int, window: int) -> float:
        transformed = ret.rolling(window).sum() if window > 1 else ret
        if len(transformed) < lag:
            return 0.0
        value = transformed.iloc[-lag]
        if pd.isna(value):
            return 0.0
        return float(np.sign(value))

    def _normalize_component(self, component: str | dict) -> dict:
        if isinstance(component, str):
            return {
                "ticker": component,
                "field": "price",
                "lag": DEFAULT_LAG,
                "window": DEFAULT_WINDOW,
            }
        return {
            "ticker": component["ticker"],
            "field": component.get("field", "price"),
            "lag": int(component.get("lag", component.get("tau", DEFAULT_LAG))),
            "window": int(component.get("window", DEFAULT_WINDOW)),
        }



"""Shared CLI helpers."""

from __future__ import annotations

import click

from abel_edge.engine.feed_loader import load_feed_frame


def build_bars_loader(cfg: dict):
    strategies = cfg.get("strategies") or []
    if not strategies:
        return None

    return _dispatch_bars_loader(cfg)


def _dispatch_bars_loader(cfg: dict):
    def _loader(**kwargs):
        config = kwargs.get("config") or {}
        profile = ((cfg.get("settings") or {}).get("data_contract") or {}).get("profile", "daily")
        feed_cfg = {
            **config,
            "name": str(config.get("name") or "primary"),
            "kind": "bars",
            "profile": profile,
        }
        if "adapter" not in feed_cfg:
            default_adapter = (
                (cfg.get("settings") or {}).get("price_data", {}).get("default_adapter")
                or (cfg.get("settings") or {}).get("price_data", {}).get("default_source", "abel")
            )
            feed_cfg["adapter"] = default_adapter
        try:
            return load_feed_frame(
                feed_cfg,
                start=kwargs.get("start"),
                end=kwargs.get("end"),
                timeframe=kwargs.get("timeframe"),
                limit=kwargs.get("limit"),
                fields=kwargs.get("fields"),
            )
        except Exception as exc:
            raise click.ClickException(str(exc)) from exc

    return _loader



from __future__ import annotations

import json

from click.testing import CliRunner

from abel_edge.cli import main
from abel_edge.research.probes import probe_graph_inputs


def _bars(rows: list[tuple[str, float, float]], symbol: str):
    import pandas as pd

    if not rows:
        return pd.DataFrame(columns=["timestamp", "symbol", "close", "volume"])
    return pd.DataFrame(
        {
            "timestamp": pd.to_datetime([item[0] for item in rows], utc=True),
            "symbol": [symbol] * len(rows),
            "close": [item[1] for item in rows],
            "volume": [item[2] for item in rows],
        }
    )


def test_probe_graph_inputs_supports_cross_calendar_and_volume(monkeypatch):
    from abel_edge.research import probes as probe_module

    def _fake_fetch_bars(*, symbols, start=None, end=None, timeframe="1d", limit=None, fields=None, config=None):
        symbol = symbols[0]
        if symbol == "TSLA":
            return _bars(
                [
                    ("2020-01-03", 100.0, 1000.0),
                    ("2020-01-06", 101.0, 1100.0),
                ],
                symbol,
            )
        if symbol == "BTCUSD":
            return _bars(
                [
                    ("2020-01-03", 9000.0, 2000.0),
                    ("2020-01-04", 9200.0, 2100.0),
                    ("2020-01-05", 9300.0, 2200.0),
                ],
                symbol,
            )
        raise AssertionError(symbol)

    monkeypatch.setattr(probe_module, "fetch_bars", _fake_fetch_bars)

    report = probe_graph_inputs(
        node_ids=["BTCUSD.price", "TSLA.volume"],
        target_node="TSLA.price",
        start="2020-01-01",
    )

    results = {item["node_id"]: item for item in report["results"]}
    btc = results["BTCUSD.price"]
    volume = results["TSLA.volume"]

    assert report["target"]["node_id"] == "TSLA.price"
    assert report["basket"]["dense_overlap_start"].startswith("2020-01-03")
    assert btc["status"] == "full_target_overlap"
    assert btc["target_overlap_days"] == 2
    assert btc["asof_preview"][-1]["value"] == 9300.0
    assert volume["runtime_field"] == "volume"
    assert volume["native_sample"][0]["value"] == 1000.0


def test_probe_data_cli_writes_json(monkeypatch, tmp_path):
    from abel_edge.research import probes as probe_module

    monkeypatch.setattr(
        probe_module,
        "fetch_bars",
        lambda **kwargs: _bars(
            [
                ("2020-01-03", 100.0, 1000.0),
                ("2020-01-06", 101.0, 1100.0),
            ],
            kwargs["symbols"][0],
        ),
    )

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        result = runner.invoke(
            main,
            [
                "probe-data",
                "--target-node",
                "TSLA.price",
                "--node-id",
                "TSLA.volume",
                "--output-json",
                "probe.json",
            ],
        )
        assert result.exit_code == 0, result.output
        assert "Graph Input Probe" in result.output
        payload = json.loads(open("probe.json", "r", encoding="utf-8").read())
        assert payload["target"]["node_id"] == "TSLA.price"
        assert payload["results"][0]["node_id"] == "TSLA.volume"



"""Credential and endpoint helpers for the optional Abel plugin."""

from __future__ import annotations

import os
from pathlib import Path

DEFAULT_AUTH_BASE_URL = "https://api.abel.ai/echo"
DEFAULT_CAP_BASE_URL = "https://cap.abel.ai/api"
CAUSAL_ABEL_ENV_BASENAMES = (".env.skill", ".env.skills", ".env")


def _global_causal_abel_skill_dirs() -> tuple[Path, ...]:
    home_dir = Path.home()
    return (
        home_dir / ".config" / "opencode" / "skills" / "causal-abel",
        home_dir / ".codex" / "skills" / "causal-abel",
    )


class MissingAbelApiKeyError(RuntimeError):
    """Raised when Abel-backed features are used without credentials."""


def _read_env_file(path: str | Path) -> dict[str, str]:
    env_path = Path(path)
    if not env_path.exists():
        return {}
    values: dict[str, str] = {}
    for raw in env_path.read_text(encoding="utf-8").splitlines():
        line = raw.strip()
        if not line or line.startswith("#") or "=" not in line:
            continue
        key, value = line.split("=", 1)
        key = key.strip()
        value = value.strip().strip('"').strip("'")
        if key and key not in values:
            values[key] = value
    return values


def _dedupe_paths(paths: list[Path]) -> list[Path]:
    seen: set[Path] = set()
    unique: list[Path] = []
    for path in paths:
        if path in seen:
            continue
        seen.add(path)
        unique.append(path)
    return unique


def _source_env_path() -> Path:
    return Path(__file__).resolve().parents[3] / ".env"


def _resolve_env_path(path: str | Path) -> Path:
    env_path = Path(path).expanduser()
    if env_path.is_absolute():
        return env_path
    return (Path.cwd() / env_path).resolve()


def _candidate_env_files(*, env_path: str | Path = ".env") -> list[Path]:
    provided = _resolve_env_path(env_path)
    source_env = _source_env_path()
    if str(env_path).strip() == ".env":
        return _dedupe_paths([source_env, provided])
    return _dedupe_paths([provided, source_env])


def _find_causal_abel_skill_dir(*, env_path: str | Path = ".env") -> Path | None:
    for env_file in _candidate_env_files(env_path=env_path):
        start_dir = env_file.parent
        for directory in (start_dir, *start_dir.parents):
            skill_dir = directory / ".agents" / "skills" / "causal-abel"
            if skill_dir.is_dir():
                return skill_dir
    return None


def _candidate_shared_auth_files(*, env_path: str | Path = ".env") -> list[Path]:
    candidates: list[Path] = []
    explicit_auth_file = (os.getenv("ABEL_AUTH_ENV_FILE") or "").strip()
    if explicit_auth_file:
        candidates.append(_resolve_env_path(explicit_auth_file))

    skill_dir = _find_causal_abel_skill_dir(env_path=env_path)
    if skill_dir is not None:
        for basename in CAUSAL_ABEL_ENV_BASENAMES:
            candidates.append(skill_dir / basename)

    for skill_dir in _global_causal_abel_skill_dirs():
        if not skill_dir.is_dir():
            continue
        for basename in CAUSAL_ABEL_ENV_BASENAMES:
            candidates.append(skill_dir / basename)

    return _dedupe_paths(candidates)


def _resolve_token_from_values(values: dict[str, str]) -> str | None:
    token = normalize_api_key(values.get("ABEL_API_KEY") or values.get("CAP_API_KEY"))
    return token or None


def _api_key_record(
    *,
    token: str | None,
    source: str,
    path: str | None = None,
) -> dict[str, str | None]:
    return {
        "api_key": token,
        "source": source,
        "path": path,
    }


def normalize_api_key(value: str | None) -> str:
    token = (value or "").strip()
    return token.removeprefix("Bearer ").strip()


def _format_env_value(value: str) -> str:
    if not value or any(char.isspace() for char in value) or "#" in value:
        escaped = value.replace("\\", "\\\\").replace('"', '\\"')
        return f'"{escaped}"'
    return value


def resolve_api_key(*, env_path: str | Path = ".env") -> str | None:
    record = resolve_api_key_record(env_path=env_path)
    return str(record["api_key"]) if record["api_key"] else None


def resolve_api_key_record(*, env_path: str | Path = ".env") -> dict[str, str | None]:
    token = normalize_api_key(
        os.getenv("ABEL_API_KEY")
        or os.getenv("CAP_API_KEY")
    )
    if token:
        return _api_key_record(token=token, source="env_var", path=None)

    for candidate_env in _candidate_env_files(env_path=env_path):
        candidate_token = _resolve_token_from_values(_read_env_file(candidate_env))
        if candidate_token:
            return _api_key_record(
                token=candidate_token,
                source="project_env",
                path=str(candidate_env),
            )

    for shared_auth_file in _candidate_shared_auth_files(env_path=env_path):
        shared_token = _resolve_token_from_values(_read_env_file(shared_auth_file))
        if shared_token:
            return _api_key_record(
                token=shared_token,
                source="shared_auth_file",
                path=str(shared_auth_file),
            )

    return _api_key_record(token=None, source="missing", path=None)


def require_api_key(*, env_path: str | Path = ".env") -> str:
    record = resolve_api_key_record(env_path=env_path)
    token = str(record["api_key"]) if record["api_key"] else None
    if token:
        return token
    shared_auth_hint = (
        "Set ABEL_AUTH_ENV_FILE to a causal-abel auth file if you store it outside the project."
    )
    skill_dir = _find_causal_abel_skill_dir(env_path=env_path)
    if skill_dir is not None:
        shared_auth_hint = (
            "`causal-abel` appears to be installed. Run "
            f"`python {skill_dir / 'scripts' / 'cap_probe.py'} auth-status --compact` to confirm auth, "
            "or set ABEL_AUTH_ENV_FILE to the exported auth file."
        )
    raise MissingAbelApiKeyError(
        "Abel API key not found. Install the `causal-abel` skill and complete its OAuth flow, "
        "or set ABEL_API_KEY or CAP_API_KEY in your environment or project .env before using Abel-backed "
        "features. abel-edge also checks ABEL_AUTH_ENV_FILE, project-local "
        ".agents/skills/causal-abel/.env.skill, and known global skill installs such as "
        "~/.config/opencode/skills/causal-abel/.env.skill or ~/.codex/skills/causal-abel/.env.skill. "
        f"{shared_auth_hint}"
    )


def resolve_cap_base_url(*, env_path: str | Path = ".env") -> str:
    configured = (os.getenv("ABEL_CAP_BASE_URL") or "").strip()
    if not configured:
        for candidate_env in _candidate_env_files(env_path=env_path):
            configured = (_read_env_file(candidate_env).get("ABEL_CAP_BASE_URL") or "").strip()
            if configured:
                break
    return (configured or DEFAULT_CAP_BASE_URL).rstrip("/")


def resolve_auth_base_url(*, env_path: str | Path = ".env") -> str:
    configured = (os.getenv("ABEL_AUTH_BASE_URL") or "").strip()
    if not configured:
        for candidate_env in _candidate_env_files(env_path=env_path):
            configured = (_read_env_file(candidate_env).get("ABEL_AUTH_BASE_URL") or "").strip()
            if configured:
                break
    return (configured or DEFAULT_AUTH_BASE_URL).rstrip("/")


def persist_env_value(*, env_path: str | Path = ".env", key: str, value: str) -> None:
    path = Path(env_path)
    path.parent.mkdir(parents=True, exist_ok=True)
    lines: list[str] = []
    if path.exists():
        lines = path.read_text(encoding="utf-8").splitlines()

    assignment = f"{key}={_format_env_value(value)}"
    for index, line in enumerate(lines):
        stripped = line.strip()
        if stripped.startswith(f"{key}="):
            lines[index] = assignment
            break
    else:
        if lines and lines[-1].strip():
            lines.append("")
        lines.append(assignment)

    path.write_text("\n".join(lines) + "\n", encoding="utf-8")



from __future__ import annotations

import pandas as pd


def filter_tracking_rows(df: pd.DataFrame | None) -> pd.DataFrame | None:
    if df is None or len(df) == 0 or "source" not in df.columns:
        return None
    tracked = df[df["source"].astype(str).str.lower() == "live"].copy()
    return tracked if len(tracked) > 0 else None


def filter_backtest_rows(df: pd.DataFrame | None) -> pd.DataFrame | None:
    if df is None or len(df) == 0:
        return None
    if "source" not in df.columns:
        return df.copy()
    backtest = df[df["source"].astype(str).str.lower() != "live"].copy()
    return backtest if len(backtest) > 0 else None


def paper_rows(df: pd.DataFrame | None, *, signal_label, fmt_pnl_pct) -> list[dict]:
    tracked = filter_tracking_rows(df)
    if tracked is None:
        return []
    rows = tracked.sort_values("date", ascending=False).head(8)
    result = []
    for _, row in rows.iterrows():
        result.append(
            {
                "date": pd.to_datetime(row["date"]).date().isoformat(),
                "close": "N/A" if pd.isna(row.get("close")) else f"{float(row['close']):.2f}",
                "signal": signal_label(float(row.get("next_position", row.get("position", 0.0)))),
                "next_position": f"{float(row.get('next_position', row.get('position', 0.0))):.2f}",
                "pnl": fmt_pnl_pct(float(row.get("pnl", 0.0))),
            }
        )
    return result



"""Tests for engine-backed research evaluation helpers."""

import json
import warnings
from pathlib import Path

from click.testing import CliRunner
import pandas as pd
import pytest

from abel_edge.cli import main
from abel_edge.engine.base import StrategyEngine
import abel_edge.research.evaluate as research_evaluate
from abel_edge.research.evaluate import (
    check_look_ahead,
    compute_k,
    render_validation_markdown,
    run_evaluation,
    run_preflight,
)
from abel_edge.research.data_readiness import run_data_verification
from abel_edge.research.workspace import init_workspace


def _write_engine(path: Path, *, bias: float = 0.02, flat: bool = False) -> None:
    if flat:
        body = (
            "        dates = pd.date_range('2024-01-01', periods=60, freq='D', tz='UTC')\n"
            "        positions = np.ones(60)\n"
            "        returns = np.full(60, 0.01)\n"
            "        prices = 100.0 * np.cumprod(1.0 + returns)\n"
        )
    else:
        body = (
            "        dates = pd.date_range('2024-01-01', periods=120, freq='D', tz='UTC')\n"
            "        phase = np.linspace(0, 8 * np.pi, 120)\n"
            "        positions = np.where(np.sin(phase) > 0, 1.0, -1.0)\n"
            f"        returns = {bias} * positions + 0.002 * np.sin(phase)\n"
            "        prices = 100.0 * np.cumprod(1.0 + returns)\n"
        )

    path.write_text(
        "\n".join(
            [
                "import numpy as np",
                "import pandas as pd",
                "",
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_signals(self):",
                *body.rstrip("\n").splitlines(),
                "        return positions, dates, prices",
                "",
                "    def get_latest_signal(self):",
                "        positions, dates, _ = self.compute_signals()",
                "        return {'position': float(positions[-1]), 'date': str(dates[-1].date())}",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_alpha_context(path: Path, *, count: int = 23) -> Path:
    path.write_text(
        json.dumps(
            {
                "validation_context": {
                    "dsr_trials": {
                        "count": count,
                        "source": "abel-invest.session/v1",
                        "method": "session_effective_exploration_trials_v1",
                        "scope": "ticker_session_requested_window",
                        "components": {
                            "prior_effective_trials": max(count - 1, 0),
                            "current_round_trials": 1,
                        },
                    }
                }
            }
        ),
        encoding="utf-8",
    )
    return path


def _write_start_aware_engine(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "import numpy as np",
                "import pandas as pd",
                "",
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_signals(self):",
                "        requested = ((self.context or {}).get('_research') or {}).get('requested_window') or {}",
                "        start = requested.get('start') or '2024-01-01'",
                "        dates = pd.date_range(start, periods=120, freq='D', tz='UTC')",
                "        phase = np.linspace(0, 8 * np.pi, 120)",
                "        positions = np.where(np.sin(phase) > 0, 1.0, -1.0)",
                "        returns = 0.02 * positions + 0.002 * np.sin(phase)",
                "        prices = 100.0 * np.cumprod(1.0 + returns)",
                "        return positions, dates, prices",
                "",
                "    def get_latest_signal(self):",
                "        positions, dates, _ = self.compute_signals()",
                "        return {'position': float(positions[-1]), 'date': str(dates[-1].date())}",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_alignment_failure_engine(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "import numpy as np",
                "import pandas as pd",
                "",
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_signals(self):",
                "        dates = pd.date_range('2024-01-01', periods=40, freq='D', tz='UTC')",
                "        aux = pd.Series(",
                "            np.arange(3, dtype=float),",
                "            index=pd.to_datetime(['2024-01-01', '2024-01-03', '2024-01-05'], utc=True),",
                "        )",
                "        self.align_series(aux, dates, method=None, allow_gaps=False)",
                "        prices = np.linspace(100.0, 120.0, len(dates))",
                "        positions = np.zeros(len(dates), dtype=float)",
                "        return positions, dates, prices",
                "",
                "    def get_latest_signal(self):",
                "        return {'position': 0.0}",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_naive_dates_engine(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "import numpy as np",
                "import pandas as pd",
                "",
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_signals(self):",
                "        dates = pd.bdate_range('2024-01-01', periods=40)",
                "        prices = np.linspace(100.0, 120.0, len(dates))",
                "        positions = np.zeros(len(dates), dtype=float)",
                "        return positions, dates, prices",
                "",
                "    def get_latest_signal(self):",
                "        return {'position': 0.0}",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_feed_loader_engine(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "import numpy as np",
                "import pandas as pd",
                "",
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_signals(self):",
                "        bars = self.load_research_bars(limit=40)",
                "        target = self.research_target_ticker() or 'SONY'",
                "        target_bars = bars[bars['symbol'] == target].copy().sort_values('timestamp')",
                "        dates = pd.DatetimeIndex(target_bars['timestamp'])",
                "        prices = target_bars['close'].astype(float).to_numpy()",
                "        positions = np.zeros(len(dates), dtype=float)",
                "        return positions, dates, prices",
                "",
                "    def get_latest_signal(self):",
                "        return {'position': 0.0}",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_decision_context_engine(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "import pandas as pd",
                "",
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_decisions(self, ctx):",
                "        close = ctx.target.series('close')",
                "        fast = close.rolling(window=3, min_periods=2).mean()",
                "        slow = close.rolling(window=5, min_periods=3).mean()",
                "        next_position = (fast > slow).astype(float).fillna(0.0)",
                "        if len(next_position) > 0:",
                "            next_position.iloc[0] = 0.0",
                "        return ctx.decisions(next_position)",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_decision_context_escape_engine(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_decisions(self, ctx):",
                "        self.research_close_frame(limit=20)",
                "        return ctx.decisions([0.0])",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_decision_context_input_engine(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "from abel_edge.engine.base import StrategyEngine",
                "",
                "",
                "class BranchEngine(StrategyEngine):",
                "    def compute_decisions(self, ctx):",
                "        driver = ctx.input('driver').asof_series('close').astype(float)",
                "        baseline = driver.rolling(window=2, min_periods=1).mean()",
                "        next_position = (driver > baseline).astype(float).fillna(0.0)",
                "        return ctx.decisions(next_position)",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_csv_context(tmp_path: Path) -> Path:
    bars_path = tmp_path / "target.csv"
    bars_path.write_text(
        "timestamp,close\n"
        "2024-01-01T00:00:00Z,100\n"
        "2024-01-02T00:00:00Z,101\n"
        "2024-01-03T00:00:00Z,103\n"
        "2024-01-04T00:00:00Z,104\n"
        "2024-01-05T00:00:00Z,106\n"
        "2024-01-06T00:00:00Z,108\n"
        "2024-01-07T00:00:00Z,109\n"
        "2024-01-08T00:00:00Z,111\n"
        "2024-01-09T00:00:00Z,113\n"
        "2024-01-10T00:00:00Z,114\n"
        "2024-01-11T00:00:00Z,116\n"
        "2024-01-12T00:00:00Z,117\n"
        "2024-01-13T00:00:00Z,119\n"
        "2024-01-14T00:00:00Z,121\n"
        "2024-01-15T00:00:00Z,122\n"
        "2024-01-16T00:00:00Z,124\n"
        "2024-01-17T00:00:00Z,125\n"
        "2024-01-18T00:00:00Z,127\n"
        "2024-01-19T00:00:00Z,128\n"
        "2024-01-20T00:00:00Z,130\n"
        "2024-01-21T00:00:00Z,131\n"
        "2024-01-22T00:00:00Z,133\n"
        "2024-01-23T00:00:00Z,135\n"
        "2024-01-24T00:00:00Z,136\n"
        "2024-01-25T00:00:00Z,138\n"
        "2024-01-26T00:00:00Z,139\n"
        "2024-01-27T00:00:00Z,141\n"
        "2024-01-28T00:00:00Z,143\n"
        "2024-01-29T00:00:00Z,144\n"
        "2024-01-30T00:00:00Z,146\n"
        "2024-01-31T00:00:00Z,147\n"
        "2024-02-01T00:00:00Z,149\n"
        "2024-02-02T00:00:00Z,151\n"
        "2024-02-03T00:00:00Z,152\n"
        "2024-02-04T00:00:00Z,154\n"
        "2024-02-05T00:00:00Z,155\n",
        encoding="utf-8",
    )
    context_path = tmp_path / "context.json"
    context_path.write_text(
        json.dumps(
            {
                "ticker": "SONY",
                "discovery": {"ticker": "SONY"},
                "_feeds": {
                    "primary": {
                        "name": "primary",
                        "kind": "bars",
                        "adapter": "csv",
                        "timeframe": "1d",
                        "symbol": "SONY",
                        "profile": "daily",
                        "path": str(bars_path),
                    }
                },
            }
        ),
        encoding="utf-8",
    )
    return context_path


def _write_prepared_input_context(tmp_path: Path) -> Path:
    context_path = _write_csv_context(tmp_path)
    payload = json.loads(context_path.read_text(encoding="utf-8"))
    driver_path = tmp_path / "driver.csv"
    driver_path.write_text(
        "timestamp,close\n"
        "2024-01-20T00:00:00Z,50\n"
        "2024-01-21T00:00:00Z,51\n"
        "2024-01-22T00:00:00Z,52\n"
        "2024-01-23T00:00:00Z,53\n"
        "2024-01-24T00:00:00Z,54\n"
        "2024-01-25T00:00:00Z,55\n"
        "2024-01-26T00:00:00Z,56\n"
        "2024-01-27T00:00:00Z,57\n"
        "2024-01-28T00:00:00Z,58\n",
        encoding="utf-8",
    )
    payload["_feeds"]["driver"] = {
        "name": "driver",
        "kind": "bars",
        "adapter": "csv",
        "timeframe": "1d",
        "symbol": "PEER",
        "profile": "daily",
        "path": str(driver_path),
    }
    payload["data_manifest"] = {"selected_inputs": [{"node_id": "driver"}]}
    payload["window_availability"] = {
        "requested_start": "2024-01-01",
        "effective_window": {
            "start": "2024-01-20T00:00:00+00:00",
            "end": "2024-01-28T00:00:00+00:00",
        },
        "per_input_coverage": [
            {"node_id": "driver", "status": "partial_target_overlap"},
        ],
    }
    context_path.write_text(json.dumps(payload), encoding="utf-8")
    return context_path


class TestInitWorkspace:
    def test_creates_engine_backed_files(self, tmp_path):
        workspace = init_workspace("SOLUSD", tmp_path / "sol")
        assert (workspace / "engine.py").exists()
        assert (workspace / "results.tsv").exists()
        assert (workspace / "memory.md").exists()
        assert (workspace / "discovery.json").exists()
        template = (workspace / "engine.py").read_text(encoding="utf-8")
        discovery = json.loads((workspace / "discovery.json").read_text(encoding="utf-8"))
        assert "compute_decisions(self, ctx)" in template
        assert "ctx.decisions(" in template
        assert discovery["target_node"] == "SOLUSD.price"

    def test_idempotent_engine_file(self, tmp_path):
        workspace = init_workspace("SOL", tmp_path / "sol")
        (workspace / "engine.py").write_text("# modified", encoding="utf-8")
        init_workspace("SOL", tmp_path / "sol")
        assert (workspace / "engine.py").read_text(encoding="utf-8") == "# modified"


class TestComputeK:
    def test_counts_tickers_and_lags(self, tmp_path):
        engine = tmp_path / "engine.py"
        engine.write_text(
            'PARENTS = [("BTCUSD", 3), ("ETHUSD", 21)]\n'
            "class BranchEngine:\n"
            "    def compute_signals(self):\n"
            "        feature.shift(3)\n"
            "        feature.shift(21)\n",
            encoding="utf-8",
        )
        k_value, tickers, lags = compute_k(engine)
        assert k_value >= 4
        assert "BTCUSD" in tickers
        assert 21 in lags

    def test_filters_market_factors(self, tmp_path):
        engine = tmp_path / "engine.py"
        engine.write_text(
            'tickers = ["BTCUSD", "SPY"]\nclass BranchEngine:\n    def compute_signals(self):\n        feature.shift(5)\n',
            encoding="utf-8",
        )
        _, tickers, _ = compute_k(engine)
        assert "SPY" not in tickers


class TestLookAhead:
    def test_static_check_catches_rolling_without_shift(self, tmp_path):
        engine = tmp_path / "engine.py"
        engine.write_text("x = ret.rolling(20).mean()\npositions = x > 0", encoding="utf-8")
        assert check_look_ahead(engine)


class TestRunEvaluation:
    def test_missing_engine_file(self, tmp_path):
        result = run_evaluation(tmp_path)
        assert result["verdict"] == "ERROR"
        assert "engine.py not found" in result["failures"][0]
        assert result["runtime_facts"]["runtime_stage"] == "load_engine"
        assert result["runtime_facts"]["workflow_status"] == "not_completed"

    def test_strategy_py_only_is_rejected(self, tmp_path):
        (tmp_path / "strategy.py").write_text("def run_strategy():\n    return None\n", encoding="utf-8")
        result = run_evaluation(tmp_path)
        assert result["verdict"] == "ERROR"
        assert "engine.py not found" in result["failures"][0]

    def test_unimplemented_engine_returns_error(self, tmp_path):
        engine = tmp_path / "engine.py"
        engine.write_text(
            "\n".join(
                [
                    "from abel_edge.engine.base import StrategyEngine",
                    "",
                    "class BranchEngine(StrategyEngine):",
                    "    def compute_signals(self):",
                    "        raise NotImplementedError('todo')",
                    "",
                    "    def get_latest_signal(self):",
                    "        return {'position': 0.0}",
                ]
            )
            + "\n",
            encoding="utf-8",
        )
        result = run_evaluation(tmp_path)
        assert result["verdict"] == "ERROR"
        assert any("failed" in failure.lower() for failure in result["failures"])

    def test_viable_engine_passes(self, tmp_path):
        _write_engine(tmp_path / "engine.py")
        result = run_evaluation(tmp_path)
        assert result["verdict"] == "PASS"
        assert result["K"] >= 1
        assert result["profile"] == "equity_daily"
        assert result["implementation_contract"] == "legacy_signal_contract"
        assert result["K_detail"]["source"] == "engine_ast"
        assert any("engine_ast fallback estimate" in item for item in result["warnings"])

    def test_context_dsr_trials_overrides_engine_ast_k(self, tmp_path):
        _write_engine(tmp_path / "engine.py")
        context_path = _write_alpha_context(tmp_path / "alpha-context.json", count=23)

        result = run_evaluation(tmp_path, context_json=context_path)

        assert result["verdict"] != "ERROR"
        assert result["K"] == 23
        assert result["metrics"]["dsr_trials_used"] == 23
        assert result["K_detail"]["source"] == "alpha_context"
        assert result["K_detail"]["engine_ast_k"] == 1
        declared = result["K_detail"]["declared_dsr_trials"]
        assert declared["count"] == 23
        assert declared["components"]["current_round_trials"] == 1
        assert not any("engine_ast fallback estimate" in item for item in result["warnings"])

    def test_preflight_context_dsr_trials_overrides_engine_ast_k(self, tmp_path):
        _write_engine(tmp_path / "engine.py")
        context_path = _write_alpha_context(tmp_path / "alpha-context.json", count=17)

        result = run_preflight(tmp_path, context_json=context_path)

        assert result["verdict"] == "PASS"
        assert result["K"] == 17
        assert result["K_detail"]["source"] == "alpha_context"
        assert result["K_detail"]["declared_dsr_trials"]["count"] == 17

    def test_context_dsr_trials_survives_semantic_error(self, tmp_path):
        engine = tmp_path / "engine.py"
        engine.write_text(
            "\n".join(
                [
                    "import numpy as np",
                    "import pandas as pd",
                    "from abel_edge.engine.base import StrategyEngine",
                    "",
                    "class BranchEngine(StrategyEngine):",
                    "    def compute_signals(self):",
                    "        dates = pd.date_range('2024-01-01', periods=10, freq='D', tz='UTC')",
                    "        positions = np.ones(10)",
                    "        prices = np.linspace(100.0, 110.0, 10)",
                    "        return positions, dates, prices",
                ]
            )
            + "\n",
            encoding="utf-8",
        )
        context_path = _write_alpha_context(tmp_path / "alpha-context.json", count=9)

        result = run_evaluation(tmp_path, context_json=context_path)

        assert result["verdict"] == "ERROR"
        assert result["diagnostics"]["runtime_stage"] == "semantic_preflight"
        assert result["K"] == 9
        assert result["K_detail"]["source"] == "alpha_context"
        assert result["context_path"] == str(context_path.resolve())

    def test_decision_context_engine_passes_and_records_trace(self, tmp_path):
        _write_decision_context_engine(tmp_path / "engine.py")
        context_path = _write_csv_context(tmp_path)

        result = run_evaluation(tmp_path, context_json=context_path)

        assert result["verdict"] != "ERROR"
        assert result["implementation_contract"] == "decision_context"
        assert result["decision_trace"]
        assert result["decision_preview"]
        assert result["semantic"]["verdict"] == "PASS"
        runtime_facts = result["runtime_facts"]
        assert runtime_facts["contract"] == "abel-edge.runtime-facts/v1"
        assert runtime_facts["runtime_stage"] == "validation"
        assert runtime_facts["workflow_status"] == "evaluation_completed"
        assert runtime_facts["read_summary"]["read_count"] >= 1

    def test_start_aware_engine_records_requested_and_effective_window(self, tmp_path):
        _write_start_aware_engine(tmp_path / "engine.py")
        result = run_evaluation(tmp_path, start="2020-01-01")
        assert result["verdict"] == "PASS"
        assert result["requested_window"] == {"start": "2020-01-01", "end": None}
        assert result["effective_window"]["start"] == "2020-01-01"
        assert result["effective_window"]["end"] == "2020-04-29"

    def test_can_persist_metric_input_csv(self, tmp_path):
        _write_engine(tmp_path / "engine.py")
        output_csv = tmp_path / "artifacts" / "metric-input.csv"

        result = run_evaluation(tmp_path, output_csv=output_csv)

        assert result["verdict"] == "PASS"
        assert output_csv.exists()
        payload = output_csv.read_text(encoding="utf-8")
        assert "date,pnl,position,asset_return" in payload

    def test_includes_runtime_diagnostics_for_constant_position(self, tmp_path):
        _write_engine(tmp_path / "engine.py", flat=True)
        with warnings.catch_warnings():
            warnings.simplefilter("error", RuntimeWarning)
            result = run_evaluation(tmp_path)

        diagnostics = result["diagnostics"]
        assert diagnostics["failure_signature"] == "constant_position"
        assert diagnostics["signal"]["position_switches"] == 0
        assert diagnostics["signal"]["unique_position_count"] == 1
        signal = result["runtime_facts"]["signal_summary"]
        assert signal["finite_days"] == signal["total_days"]
        assert signal["nonfinite_days"] == 0
        assert signal["nonzero_position_ratio"] == 1.0

    def test_metric_failure_facts_are_structured_without_advice(self):
        facts = research_evaluate._metric_failure_facts(
            {
                "profile": "equity_daily",
                "metrics": {"total_return": -0.12},
                "failures": [
                    "T15 MaxDD 35.0% > 20%",
                    "PositionIC 0.000 < 0.02",
                    "PositionIC stab 19% < 55%",
                ],
            }
        )

        assert facts == [
            {
                "metric": "max_dd",
                "observed": 0.35,
                "threshold": 0.2,
                "comparison": "abs_gt",
                "profile": "equity_daily",
                "message": "T15 MaxDD 35.0% > 20%",
            },
            {
                "metric": "position_ic",
                "observed": 0.0,
                "threshold": 0.02,
                "comparison": "lt",
                "profile": "equity_daily",
                "message": "PositionIC 0.000 < 0.02",
            },
            {
                "metric": "position_ic_stability",
                "observed": 0.19,
                "threshold": 0.55,
                "comparison": "lt",
                "profile": "equity_daily",
                "message": "PositionIC stab 19% < 55%",
            },
        ]

    def test_validation_exception_surfaces_runtime_stage(self, tmp_path, monkeypatch):
        _write_engine(tmp_path / "engine.py", flat=True)

        def boom(*args, **kwargs):
            raise RuntimeError("validation exploded")

        monkeypatch.setattr(research_evaluate, "validate_strategy", boom)

        result = run_evaluation(tmp_path)

        assert result["verdict"] == "ERROR"
        diagnostics = result["diagnostics"]
        assert diagnostics["runtime_stage"] == "validation"
        assert diagnostics["failure_signature"] == "validation_failed"
        assert "validation exploded" in result["failures"][0]

    def test_reports_alignment_collapse_when_series_cannot_align(self, tmp_path):
        _write_alignment_failure_engine(tmp_path / "engine.py")
        result = run_evaluation(tmp_path)

        diagnostics = result["diagnostics"]
        assert result["verdict"] == "ERROR"
        assert diagnostics["failure_signature"] == "alignment_collapse"
        assert diagnostics["runtime_stage"] == "compute_strategy"

    def test_reports_datetime_contract_violation_for_naive_dates(self, tmp_path):
        _write_naive_dates_engine(tmp_path / "engine.py")
        result = run_evaluation(tmp_path)

        diagnostics = result["diagnostics"]
        assert result["verdict"] == "ERROR"
        assert diagnostics["failure_signature"] == "datetime_contract_violation"
        assert diagnostics["runtime_stage"] == "compute_strategy"
        assert "UTC-aware" in result["failures"][0]

    def test_decision_context_blocks_raw_strategy_escape_hatch(self, tmp_path):
        _write_decision_context_escape_engine(tmp_path / "engine.py")
        context_path = _write_csv_context(tmp_path)

        result = run_evaluation(tmp_path, context_json=context_path)

        assert result["verdict"] == "ERROR"
        assert result["diagnostics"]["failure_signature"] == "decision_context_escape_hatch"

    def test_rejects_engine_that_only_reexports_imported_class(self, tmp_path):
        helper = tmp_path / "helper_engine.py"
        helper.write_text(
            "\n".join(
                [
                    "from abel_edge.engine.base import StrategyEngine",
                    "",
                    "class HelperEngine(StrategyEngine):",
                    "    def compute_signals(self):",
                    "        raise NotImplementedError",
                    "",
                    "    def get_latest_signal(self):",
                    "        return {'position': 0.0}",
                ]
            )
            + "\n",
            encoding="utf-8",
        )
        (tmp_path / "engine.py").write_text(
            "from helper_engine import HelperEngine\n",
            encoding="utf-8",
        )

        result = run_evaluation(tmp_path)

        assert result["verdict"] == "ERROR"
        assert "module-owned StrategyEngine subclass" in result["failures"][0]

    def test_context_json_build_synthesizes_primary_feed(self, tmp_path):
        context_path = tmp_path / "context.json"
        context_path.write_text(
            json.dumps({"ticker": "SONY", "discovery": {"ticker": "SONY"}}),
            encoding="utf-8",
        )
        context = research_evaluate._build_research_context(
            workspace=tmp_path,
            start=None,
            context_json=context_path,
        )

        assert context["_feeds"]["primary"]["symbol"] == "SONY"
        assert context["_feeds"]["primary"]["adapter"] == "abel"
        assert context["_runtime_profile"]["target"] == "SONY"
        assert context["_execution_constraints"]["long_only"] is False

    def test_context_json_preserves_injected_grandma_runtime_contract(self, tmp_path):
        context_path = tmp_path / "context.json"
        context_path.write_text(
            json.dumps(
                {
                    "ticker": "TSLA",
                    "_runtime_profile": {
                        "profile": "daily",
                        "target": "TSLA",
                        "validation_profile": "grandma_daily",
                        "execution_delay_bars": 2,
                    },
                    "_execution_constraints": {
                        "long_only": False,
                        "position_bounds": [-1.0, 1.0],
                    },
                }
            ),
            encoding="utf-8",
        )

        context = research_evaluate._build_research_context(
            workspace=tmp_path,
            start=None,
            context_json=context_path,
        )

        assert context["_runtime_profile"]["validation_profile"] == "grandma_daily"
        assert context["_runtime_profile"]["execution_delay_bars"] == 2
        assert context["_execution_constraints"]["position_bounds"] == [-1.0, 1.0]

    def test_evaluation_uses_context_validation_profile(self, tmp_path):
        _write_engine(tmp_path / "engine.py")
        context_path = tmp_path / "context.json"
        context_path.write_text(
            json.dumps({"validation_context": {"profile": "grandma_daily"}}),
            encoding="utf-8",
        )

        result = run_evaluation(tmp_path, context_json=context_path)

        assert result["verdict"] == "PASS"
        assert result["profile"] == "grandma_daily"

    def test_preflight_keeps_static_checks_as_warnings(self, tmp_path):
        engine = tmp_path / "engine.py"
        engine.write_text(
            "\n".join(
                [
                    "import numpy as np",
                    "import pandas as pd",
                    "",
                    "from abel_edge.engine.base import StrategyEngine",
                    "",
                    "",
                    "class BranchEngine(StrategyEngine):",
                    "    def compute_signals(self):",
                    "        dates = pd.date_range('2024-01-01', periods=40, freq='D', tz='UTC')",
                    "        prices = np.linspace(100.0, 130.0, len(dates))",
                    "        rolling = pd.Series(prices).rolling(5).mean()",
                    "        positions = (rolling > 110).fillna(0.0).to_numpy(dtype=float)",
                    "        return positions, dates, prices",
                ]
            )
            + "\n",
            encoding="utf-8",
        )

        result = run_preflight(tmp_path)

        assert result["verdict"] == "PASS"
        assert any("Static look-ahead heuristics" in warning for warning in result["semantic"]["warnings"])

    def test_preflight_surfaces_prepared_input_window_issues(self, tmp_path):
        _write_decision_context_input_engine(tmp_path / "engine.py")
        context_path = _write_prepared_input_context(tmp_path)

        result = run_preflight(tmp_path, context_json=context_path)

        assert result["verdict"] == "PASS"
        assert result["diagnostics"]["failure_signature"] == "effective_window_collapse"
        assert "driver" in result["semantic"]["prepared_inputs"]["traced_inputs"]
        issue_kinds = [item["kind"] for item in result["semantic"]["prepared_inputs"]["issues"]]
        assert "effective_window_collapse" in issue_kinds
        assert "stale_input_tail" in issue_kinds
        assert any(
            "Prepared input availability starts after" in warning
            for warning in result["semantic"]["warnings"]
        )
        runtime_facts = result["runtime_facts"]
        assert runtime_facts["runtime_stage"] == "semantic_preflight"
        assert runtime_facts["read_summary"]["auxiliary_reads"] == ["driver"]
        assert runtime_facts["prepared_inputs"]["selected_inputs"] == ["DRIVER"]
        assert "effective_window_collapse" in runtime_facts["temporal_visibility"]["issue_kinds"]

    def test_prepared_input_selection_uses_branch_and_manifest_inputs(self):
        selected_inputs = research_evaluate._context_selected_inputs(
            {
                "branch_spec": {
                    "selected_inputs": [
                        "IMOUSD",
                        {"node_id": "SFIUSD.price", "symbol": "SFIUSD"},
                        "IMOUSD",
                    ],
                },
                "data_manifest": {
                    "selected_inputs": ["SFIUSD"],
                    "feeds": [
                        {"name": "primary", "symbol": "TSLA", "role": "target"},
                        {"name": "IMOUSD", "symbol": "IMOUSD", "role": "driver"},
                        {"name": "SFIUSD", "symbol": "SFIUSD", "role": "driver"},
                    ],
                },
            },
            target="TSLA",
        )

        assert selected_inputs == ["IMOUSD", "SFIUSD"]

    @pytest.mark.parametrize(
        "example_dir",
        [
            "sma_crossover",
            "momentum_ml",
            "causal_demo",
        ],
    )
    def test_shipped_examples_do_not_fail_datetime_contract(self, example_dir):
        root = Path(__file__).resolve().parents[1]
        result = run_evaluation(root / "examples" / example_dir)

        assert result["diagnostics"]["failure_signature"] != "datetime_contract_violation"
        assert not any("UTC-aware" in failure for failure in result["failures"])


class TestResearchHelpers:
    def test_strategy_engine_exposes_discovery_and_readiness_helpers(self):
        class DemoEngine(StrategyEngine):
            def compute_signals(self):
                raise NotImplementedError

            def get_latest_signal(self):
                return {"position": 0.0}

        engine = DemoEngine(
            context={
                "ticker": "sony",
                "discovery": {
                    "ticker": "SONY",
                    "parents": [{"ticker": "AAPL", "field": "price"}],
                    "blanket_new": [{"ticker": "MSFT", "field": "price", "roles": ["spouse"]}],
                    "children": [{"ticker": "NVDA", "field": "price"}],
                    "data_readiness": {
                        "results": [
                            {
                                "ticker": "SONY",
                                "status": "start_covered",
                                "usable": True,
                                "covers_requested_start": True,
                            },
                            {
                                "ticker": "AAPL",
                                "status": "start_covered",
                                "usable": True,
                                "covers_requested_start": True,
                                "rows": 252,
                            },
                            {
                                "ticker": "MSFT",
                                "status": "partial_window",
                                "usable": True,
                                "covers_requested_start": False,
                                "rows": 120,
                            },
                            {
                                "ticker": "NVDA",
                                "status": "no_data",
                                "usable": False,
                                "covers_requested_start": False,
                            },
                        ]
                    },
                },
                "_research": {
                    "requested_window": {"start": "2020-01-01", "end": None},
                },
            }
        )

        assert engine.research_target_ticker() == "SONY"
        assert engine.research_requested_start() == "2020-01-01"
        assert engine.research_driver_tickers() == ["AAPL", "MSFT"]
        assert engine.research_driver_tickers(require_full_window=True) == ["AAPL"]
        assert engine.research_driver_tickers(roles=("parent",)) == ["AAPL"]

        candidates = engine.research_driver_candidates(require_usable=False)
        assert [item["ticker"] for item in candidates] == ["AAPL", "MSFT", "NVDA"]
        assert candidates[0]["discovery_roles"] == ["parent"]
        assert candidates[1]["discovery_roles"] == ["blanket", "spouse"]
        assert candidates[2]["readiness_status"] == "no_data"

    def test_strategy_engine_can_load_research_bars_and_close_frame(self):
        class DemoEngine(StrategyEngine):
            def compute_signals(self):
                raise NotImplementedError

            def get_latest_signal(self):
                return {"position": 0.0}

            def _load_research_symbol_bars(self, symbol, **kwargs):
                assert kwargs["start"] == "2020-01-01"
                assert kwargs["limit"] == 600
                rows = {
                    "SONY": [
                        ("2020-01-01", 100.0),
                        ("2020-01-02", 101.0),
                    ],
                    "AAPL": [
                        ("2020-01-01", 50.0),
                        ("2020-01-02", 51.0),
                    ],
                }[symbol]
                return pd.DataFrame(
                    {
                        "timestamp": pd.to_datetime([item[0] for item in rows], utc=True),
                        "symbol": [symbol] * len(rows),
                        "close": [item[1] for item in rows],
                    }
                )

        engine = DemoEngine(
            context={
                "discovery": {
                    "ticker": "SONY",
                    "parents": [{"ticker": "AAPL", "field": "price"}],
                    "data_readiness": {
                        "results": [
                            {"ticker": "AAPL", "status": "start_covered", "usable": True, "covers_requested_start": True},
                        ]
                    },
                },
                "_research": {"requested_window": {"start": "2020-01-01", "end": None}},
            }
        )

        bars = engine.load_research_bars(require_full_window=True)
        frame = engine.research_close_frame(require_full_window=True)

        assert list(bars["symbol"]) == ["SONY", "AAPL", "SONY", "AAPL"]
        assert list(frame.columns) == ["SONY", "AAPL"]
        assert float(frame.iloc[-1]["SONY"]) == 101.0
        assert float(frame.iloc[-1]["AAPL"]) == 51.0

    def test_same_asset_volume_candidate_survives_target_exclusion(self):
        class DemoEngine(StrategyEngine):
            def compute_signals(self):
                raise NotImplementedError

            def get_latest_signal(self):
                return {"position": 0.0}

        engine = DemoEngine(
            context={
                "discovery": {
                    "ticker": "SONY",
                    "target_node": "SONY.price",
                    "parents": [],
                    "blanket_new": [{"ticker": "SONY", "field": "volume", "roles": ["sibling"]}],
                    "children": [],
                    "data_readiness": {
                        "results": [
                            {
                                "ticker": "SONY",
                                "status": "start_covered",
                                "usable": True,
                                "covers_requested_start": True,
                            }
                        ]
                    },
                }
            }
        )

        candidates = engine.research_driver_candidates(require_usable=False)

        assert len(candidates) == 1
        assert candidates[0]["node_id"] == "SONY.volume"
        assert candidates[0]["ticker"] == "SONY"

    def test_load_research_bars_rejects_empty_symbol_selection(self):
        class DemoEngine(StrategyEngine):
            def compute_signals(self):
                raise NotImplementedError

            def get_latest_signal(self):
                return {"position": 0.0}

        engine = DemoEngine(context={"_research": {"requested_window": {"start": "2020-01-01"}}})

        with pytest.raises(ValueError, match="No research symbols were selected"):
            engine.load_research_bars(include_target=False, driver_tickers=[])

    def test_strategy_engine_can_build_target_driver_frame_with_safe_overlap_modes(self):
        class DemoEngine(StrategyEngine):
            def compute_signals(self):
                raise NotImplementedError

            def get_latest_signal(self):
                return {"position": 0.0}

            def _load_research_symbol_bars(self, symbol, **kwargs):
                rows = {
                    "SONY": [
                        ("2020-01-01", 100.0),
                        ("2020-01-02", 101.0),
                        ("2020-01-03", 102.0),
                    ],
                    "AAPL": [
                        ("2020-01-01", 50.0),
                        ("2020-01-03", 52.0),
                    ],
                    "MSFT": [
                        ("2020-01-01", 20.0),
                        ("2020-01-02", 21.0),
                        ("2020-01-03", 22.0),
                    ],
                }[symbol]
                return pd.DataFrame(
                    {
                        "timestamp": pd.to_datetime([item[0] for item in rows], utc=True),
                        "symbol": [symbol] * len(rows),
                        "close": [item[1] for item in rows],
                    }
                )

        engine = DemoEngine(
            context={
                "discovery": {
                    "ticker": "SONY",
                    "parents": [{"ticker": "AAPL"}, {"ticker": "MSFT"}],
                    "data_readiness": {
                        "results": [
                            {"ticker": "AAPL", "status": "start_covered", "usable": True, "covers_requested_start": True},
                            {"ticker": "MSFT", "status": "start_covered", "usable": True, "covers_requested_start": True},
                        ]
                    },
                },
                "_research": {"requested_window": {"start": "2020-01-01", "end": None}},
            }
        )

        target_intersection, drivers_intersection = engine.research_target_driver_frame(
            require_full_window=True,
            overlap="intersection",
            require_drivers=True,
        )
        target_target_only, drivers_target_only = engine.research_target_driver_frame(
            require_full_window=True,
            overlap="target_only",
            require_drivers=True,
        )

        assert list(target_intersection.index.strftime("%Y-%m-%d")) == ["2020-01-01", "2020-01-03"]
        assert list(drivers_intersection.columns) == ["AAPL", "MSFT"]
        assert list(target_target_only.index.strftime("%Y-%m-%d")) == [
            "2020-01-01",
            "2020-01-02",
            "2020-01-03",
        ]
        assert pd.isna(drivers_target_only.loc[pd.Timestamp("2020-01-02", tz="UTC"), "AAPL"])


class TestValidationMarkdown:
    def test_renders_validation_summary(self, tmp_path):
        _write_engine(tmp_path / "engine.py")
        result = run_evaluation(tmp_path)
        report = render_validation_markdown(result)
        assert "# Evaluation Summary" in report
        assert "## Verdict" in report
        assert "## Prepared Inputs" in report
        assert result["verdict"] in report
        assert "implementation_contract" in report


class TestEvaluateCli:
    def test_evaluate_cli_writes_raw_outputs(self, tmp_path):
        runner = CliRunner()
        with runner.isolated_filesystem(temp_dir=tmp_path):
            workdir = Path("workspace")
            workdir.mkdir()
            _write_engine(workdir / "engine.py")

            result = runner.invoke(
                main,
                [
                    "evaluate",
                    "--workdir",
                    str(workdir),
                    "--output-json",
                    str(workdir / "edge-result.json"),
                    "--output-md",
                    str(workdir / "edge-validation.md"),
                    "--output-csv",
                    str(workdir / "metric-input.csv"),
                ],
            )
            assert result.exit_code == 0, result.output
            assert (workdir / "edge-result.json").exists()
            assert (workdir / "edge-validation.md").exists()
            assert (workdir / "metric-input.csv").exists()
            assert "Verdict: PASS" in result.output
            assert "Input CSV:" in result.output

    def test_evaluate_cli_passes_start_into_research_context(self, tmp_path):
        runner = CliRunner()
        with runner.isolated_filesystem(temp_dir=tmp_path):
            workdir = Path("workspace")
            workdir.mkdir()
            _write_start_aware_engine(workdir / "engine.py")

            result = runner.invoke(
                main,
                [
                    "evaluate",
                    "--workdir",
                    str(workdir),
                    "--start",
                    "2020-01-01",
                    "--output-json",
                    str(workdir / "edge-result.json"),
                ],
            )
            assert result.exit_code == 0, result.output
            payload = (workdir / "edge-result.json").read_text(encoding="utf-8")
            assert '"start": "2020-01-01"' in payload

    def test_evaluate_cli_fails_for_bad_engine(self, tmp_path):
        runner = CliRunner()
        with runner.isolated_filesystem(temp_dir=tmp_path):
            workdir = Path("workspace")
            workdir.mkdir()
            _write_engine(workdir / "engine.py", flat=True)

            result = runner.invoke(main, ["evaluate", "--workdir", str(workdir)])
            assert result.exit_code != 0
            assert "Verdict: FAIL" in result.output

    def test_debug_evaluate_cli_surfaces_failure_signature(self, tmp_path):
        runner = CliRunner()
        with runner.isolated_filesystem(temp_dir=tmp_path):
            workdir = Path("workspace")
            workdir.mkdir()
            _write_engine(workdir / "engine.py", flat=True)

            result = runner.invoke(main, ["debug-evaluate", "--workdir", str(workdir)])
            assert result.exit_code == 0, result.output
            assert "Failure signature: constant_position" in result.output
            assert "Signal activity:" in result.output


class TestVerifyData:
    def test_run_data_verification_reports_full_partial_and_missing(self, monkeypatch, tmp_path):
        from abel_edge.research import data_readiness as data_module

        def _fake_fetch_bars(*, symbols, start=None, end=None, timeframe="1d", limit=None, fields=None, config=None):
            ticker = symbols[0]
            if ticker == "SONY":
                return _bars_frame(["2020-01-01", "2020-01-02", "2020-01-03"])
            if ticker == "CNET":
                return _bars_frame(["2020-08-01", "2020-08-02"])
            if ticker == "EMPTY":
                return _bars_frame([])
            raise RuntimeError("boom")

        monkeypatch.setattr(data_module, "fetch_bars", _fake_fetch_bars)
        discovery = tmp_path / "discovery.json"
        discovery.write_text(
            json.dumps(
                {
                    "ticker": "SONY",
                    "backtest": {"start": "2020-01-01"},
                    "parents": [{"ticker": "CNET"}],
                    "blanket_new": [{"ticker": "EMPTY"}],
                    "children": [{"ticker": "BROKEN"}],
                }
            ),
            encoding="utf-8",
        )

        report = run_data_verification(discovery_json=discovery)

        summary = report["summary"]
        assert summary["ticker_count"] == 4
        assert summary["start_covered_count"] == 1
        assert summary["partial_window_count"] == 1
        assert summary["no_data_count"] == 1
        assert summary["error_count"] == 1
        assert report["probe"]["limit"] == 500
        assert report["probe"]["strategy"] == "target_boundary_confirm"
        assert report["target_boundary"]["classification"] == "confirmed_before_requested_start"
        assert report["target_boundary"]["observed_first_timestamp"] == "2020-01-01"
        assert report["coverage_hints"]["target_safe_start"] == "2020-01-01"
        assert report["coverage_hints"]["dense_overlap_hint_start"] == "2020-08-01"
        cnet = next(item for item in report["results"] if item["ticker"] == "CNET")
        assert cnet["observed_first_timestamp"] == "2020-08-01"
        assert cnet["covers_requested_start"] is False
        assert cnet["left_boundary_confidence"] == "confirmed"

    def test_run_data_verification_marks_target_boundary_unknown_when_probe_truncates(self, monkeypatch, tmp_path):
        from abel_edge.research import data_readiness as data_module

        def _fake_fetch_bars(*, symbols, start=None, end=None, timeframe="1d", limit=None, fields=None, config=None):
            ticker = symbols[0]
            if ticker != "META":
                return _bars_frame(["2024-11-25", "2024-11-26"])
            if limit == 500:
                return _bars_range("2024-04-09", periods=500)
            if limit == 1000:
                return _bars_range("2023-04-10", periods=1000)
            return _bars_range("2022-04-11", periods=2000)

        monkeypatch.setattr(data_module, "fetch_bars", _fake_fetch_bars)
        discovery = tmp_path / "discovery.json"
        discovery.write_text(
            json.dumps(
                {
                    "ticker": "META",
                    "backtest": {"start": "2020-01-01"},
                    "parents": [{"ticker": "PEER"}],
                }
            ),
            encoding="utf-8",
        )

        report = run_data_verification(discovery_json=discovery)

        assert report["target_boundary"]["classification"] == "unknown_probe_truncated"
        assert report["target_boundary"]["observed_first_timestamp"] == "2022-04-11"
        assert report["probe"]["target_confirmation_attempted"] is True
        assert report["probe"]["target_final_limit"] == 2000
        assert report["coverage_hints"]["target_safe_start"] == "2022-04-11"
        meta = next(item for item in report["results"] if item["ticker"] == "META")
        assert meta["left_boundary_confidence"] == "observed"
        assert meta["covers_requested_start"] is False

    def test_verify_data_cli_writes_json(self, monkeypatch, tmp_path):
        from abel_edge.research import data_readiness as data_module

        monkeypatch.setattr(
            data_module,
            "fetch_bars",
            lambda **kwargs: _bars_frame(["2020-01-01", "2020-01-02"]),
        )

        runner = CliRunner()
        with runner.isolated_filesystem(temp_dir=tmp_path):
            output_json = Path("report.json")
            result = runner.invoke(
                main,
                [
                    "verify-data",
                    "--ticker",
                    "SONY",
                    "--start",
                    "2020-01-01",
                    "--output-json",
                    str(output_json),
                ],
            )
            assert result.exit_code == 0, result.output
            assert "Research Data Verification" in result.output
            assert output_json.exists()
            payload = json.loads(output_json.read_text(encoding="utf-8"))
            assert payload["summary"]["usable_count"] == 1


def _bars_frame(dates: list[str]):
    import pandas as pd

    if not dates:
        return pd.DataFrame(columns=["timestamp", "symbol", "close"])
    return pd.DataFrame(
        {
            "timestamp": pd.to_datetime(dates, utc=True),
            "symbol": ["SONY"] * len(dates),
            "close": [100.0 + idx for idx in range(len(dates))],
        }
    )


def _bars_range(start: str, periods: int):
    import pandas as pd

    dates = pd.date_range(start, periods=periods, freq="D", tz="UTC")
    return pd.DataFrame(
        {
            "timestamp": dates,
            "symbol": ["META"] * len(dates),
            "close": [100.0 + idx for idx in range(len(dates))],
        }
    )



"""Look-ahead bias checks: static heuristics plus runtime leak detection."""

from __future__ import annotations

import ast
import re
from pathlib import Path

import numpy as np

from abel_edge.validation._look_ahead_ast import (
    collect_scope_bindings,
    in_string_literal,
    is_bounded_expr,
    is_numpy_reduction,
    node_offset,
    numpy_call_name,
    safe_unparse,
    string_literal_spans,
)


def check_static(source: str) -> list[str]:
    """Run static T2-T5 look-ahead checks against Python source."""
    lines = source.split("\n")
    raw: list[str] = []
    raw.extend(_t2_rolling_without_shift(source))
    raw.extend(_t3_global_stats(source))
    raw.extend(_t4_walk_forward_slicing(source))
    raw.extend(_t5_trend_filter(source))
    return [violation for violation in raw if not _is_suppressed(violation, lines)]


def check_static_file(path: str | Path) -> list[str]:
    return check_static(Path(path).read_text(encoding="utf-8"))


def check_runtime(
    pnl: np.ndarray,
    positions: np.ndarray,
    returns: np.ndarray | None = None,
    *,
    threshold: float = 0.3,
    hit_rate_max: float = 0.70,
) -> list[str]:
    """Run runtime leak checks against aligned pnl/position/return arrays."""
    warnings: list[str] = []
    active = np.abs(positions) > 0.01

    if returns is not None and len(returns) == len(positions):
        if active.sum() > 30:
            pos_mag = np.abs(positions[active])
            ret_mag = np.abs(returns[active])
            if np.std(pos_mag) > 1e-12 and np.std(ret_mag) > 1e-12:
                corr = np.corrcoef(pos_mag, ret_mag)[0, 1]
                if not np.isnan(corr) and abs(corr) > threshold:
                    warnings.append(
                        f"R1: |position| correlates with |same-day return| "
                        f"(corr={corr:.3f}, threshold={threshold}). "
                        f"Positions may be using future information."
                    )

    active_pnl = pnl[active]
    if len(active_pnl) > 30:
        reference = active_pnl
        if returns is not None and len(returns) == len(positions):
            reference = returns[active]
        if np.std(reference) > 1e-12:
            hit_rate = float(np.mean(active_pnl > 0))
            if hit_rate > hit_rate_max:
                warnings.append(
                    f"R2: Hit rate {hit_rate:.0%} on {len(active_pnl)} active days "
                    f"exceeds {hit_rate_max:.0%}. Verify no look-ahead in features."
                )

    return warnings


def _is_suppressed(violation: str, lines: list[str]) -> bool:
    match = re.match(r"^(T[2-5])\s+L(\d+):", violation)
    if not match:
        return False

    code = match.group(1).lower()
    line_num = int(match.group(2))
    candidates: list[str] = []
    if 0 <= line_num - 1 < len(lines):
        candidates.append(lines[line_num - 1])
    if 0 <= line_num - 2 < len(lines):
        candidates.append(lines[line_num - 2])

    for line in candidates:
        noqa = re.search(r"#\s*noqa(?::\s*([A-Za-z0-9_,\s]+))?", line)
        if not noqa:
            continue
        arg = noqa.group(1)
        if arg is None:
            return True
        tokens = {token.strip().lower() for token in arg.split(",")}
        if code in tokens or "lookahead" in tokens:
            return True
    return False


def _t2_rolling_without_shift(source: str) -> list[str]:
    violations: list[str] = []
    spans = string_literal_spans(source)
    pattern = re.compile(
        r"\.rolling\(\s*\d+\s*\)\s*\.\s*(mean|std|sum|var|median|corr|min|max)\s*\([^)]*\)"
    )
    for match in pattern.finditer(source):
        if in_string_literal(match.start(), spans):
            continue
        line_num = source[: match.start()].count("\n") + 1
        after = source[match.end() : match.end() + 50]
        if ".shift(" in after:
            continue
        ctx = source[max(0, match.start() - 30) : match.end() + 50]
        if "shift(-" in ctx or "# target" in ctx.lower() or "# label" in ctx.lower():
            continue
        line_start = source.rfind("\n", 0, match.start()) + 1
        line = source[line_start : match.end() + 50]
        if re.match(r"\s*(y|target|label)\s*=", line):
            continue
        snippet = source[max(0, match.start() - 5) : match.end() + 20].strip()
        violations.append(
            f"T2 L{line_num}: rolling().{match.group(1)}() without .shift(1): {snippet}"
        )
    return violations


def _t3_global_stats(source: str) -> list[str]:
    try:
        tree = ast.parse(source)
    except SyntaxError:
        return []

    violations: list[str] = []
    spans = string_literal_spans(source)
    scope_cache: dict[int, dict[str, ast.AST]] = {}

    def _scan(node: ast.AST, scope: ast.AST) -> None:
        if isinstance(node, ast.Call) and is_numpy_reduction(node) and node.args:
            offset = node_offset(node, source)
            if offset is None or not in_string_literal(offset, spans):
                scope_id = id(scope)
                if scope_id not in scope_cache:
                    scope_cache[scope_id] = collect_scope_bindings(scope)
                if not is_bounded_expr(node.args[0], scope_cache[scope_id]):
                    violations.append(
                        f"T3 L{node.lineno}: "
                        f"np.{numpy_call_name(node)}({safe_unparse(node.args[0])}) "
                        f"on full array - use rolling/expanding or [:i] slice"
                    )
        for child in ast.iter_child_nodes(node):
            new_scope = child if isinstance(child, (ast.FunctionDef, ast.AsyncFunctionDef)) else scope
            _scan(child, new_scope)

    _scan(tree, tree)
    return violations


def _t4_walk_forward_slicing(source: str) -> list[str]:
    violations: list[str] = []
    pattern = re.compile(r"\[\s*:?\s*(\w+)\s*\+\s*1\s*\]")
    for match in pattern.finditer(source):
        line_num = source[: match.start()].count("\n") + 1
        line_start = source.rfind("\n", 0, match.start()) + 1
        line_end = source.find("\n", match.end())
        if line_end == -1:
            line_end = len(source)
        line = source[line_start:line_end]
        if any(keyword in line.lower() for keyword in ("train", "fit", "x_tr", "y_tr", "x_s", "weight")):
            violations.append(
                f"T4 L{line_num}: [:i+1] in training context - use [:i] to exclude current day"
            )
    return violations


def _t5_trend_filter(source: str) -> list[str]:
    violations: list[str] = []
    pattern = re.compile(
        r"(close|price|px|eth_price|sma|ma)\s*\[\s*(\w+)\s*\]"
        r"\s*[<>]=?\s*"
        r"(close|price|px|sma|ma)\s*\[\s*(\w+)\s*\]"
    )
    for match in pattern.finditer(source):
        idx_left, idx_right = match.group(2), match.group(4)
        if idx_left != idx_right or "-1" in idx_left:
            continue
        line_num = source[: match.start()].count("\n") + 1
        line_start = source.rfind("\n", 0, match.start()) + 1
        line_end = source.find("\n", match.end())
        if line_end == -1:
            line_end = len(source)
        line = source[line_start:line_end]
        if "shift(1)" in line or "i-1" in line or "i - 1" in line:
            continue
        violations.append(
            f"T5 L{line_num}: trend filter uses current-day index - use [i-1] or .shift(1)"
        )
    return violations



"""Formatting helpers for dashboard rendering."""

from __future__ import annotations


def fmt_pnl_pct(value: float) -> str:
    """Format PnL as percentage string."""
    return f"{value * 100:+.1f}%"


def fmt_dollar(value: float) -> str:
    """Format dollar amount."""
    return f"${value:,.0f}"



"""Package metadata consistency tests."""

from pathlib import Path
import tomllib

import abel_edge


ROOT = Path(__file__).resolve().parent.parent


def test_runtime_version_matches_project_metadata() -> None:
    data = tomllib.loads((ROOT / "pyproject.toml").read_text(encoding="utf-8"))

    assert abel_edge.__version__ == data["project"]["version"]



"""Small Abel CAP client used by the optional plugin."""

from __future__ import annotations

import json
from pathlib import Path
import shutil
import subprocess
import sys
import time
from typing import Any
import uuid

import requests

from abel_edge.plugins.abel.credentials import resolve_cap_base_url

CAP_VERSION = "0.2.2"
DEFAULT_GRAPH_ID = "abel-main"
DEFAULT_GRAPH_VERSION = "CausalNodeV3"
DEFAULT_TIMEOUT_SECONDS = 60
DEFAULT_RETRY_ATTEMPTS = 4

CRYPTO_ALIASES = {"BTC", "ETH", "SOL", "XRP", "DOGE", "ADA", "AVAX"}
SUPPORTED_FIELDS = {"price", "volume"}
SUPPORTED_MARKET_FIELDS = {"open", "high", "low", "close", "volume"}


def normalize_public_node_id(value: str, *, default_field: str = "price") -> str:
    raw = value.strip()
    if not raw:
        raise ValueError("Ticker or node id cannot be empty.")
    if default_field not in SUPPORTED_FIELDS:
        raise ValueError(f"Unsupported field '{default_field}'.")

    normalized = raw.upper()
    ticker, dot, suffix = normalized.rpartition(".")
    if dot:
        field = suffix.lower()
        if field not in SUPPORTED_FIELDS:
            raise ValueError("Abel node ids must end with .price or .volume.")
        return f"{ticker}.{field}"

    ticker, underscore, suffix = normalized.rpartition("_")
    if underscore:
        field = suffix.lower()
        if field == "close":
            return f"{ticker}.price"
        if field == "volume":
            return f"{ticker}.volume"
        raise ValueError("Abel node ids must use .price or .volume.")

    if normalized in CRYPTO_ALIASES:
        normalized = f"{normalized}USD"
    return f"{normalized}.{default_field}"


def split_public_node_id(node_id: str) -> tuple[str, str]:
    ticker, _, field = normalize_public_node_id(node_id).rpartition(".")
    return ticker, field


def _serialize_timestamp(value):
    if value is None:
        return None
    if hasattr(value, "isoformat"):
        return value.isoformat()
    return str(value)


def _normalize_market_fields(fields: list[str] | None) -> list[str]:
    requested = fields or ["open", "high", "low", "close", "volume"]
    normalized = []
    seen = set()
    for field in requested:
        name = str(field).strip().lower()
        if name in {"timestamp", "symbol", "date"}:
            continue
        if name not in SUPPORTED_MARKET_FIELDS:
            continue
        if name not in seen:
            seen.add(name)
            normalized.append(name)
    if not normalized:
        return ["open", "high", "low", "close", "volume"]
    return normalized


class AbelClient:
    def __init__(
        self,
        *,
        cap_base_url: str | None = None,
        env_path: str | Path = ".env",
        session: requests.Session | None = None,
    ) -> None:
        self.cap_base_url = (cap_base_url or resolve_cap_base_url(env_path=env_path)).rstrip("/")
        self.session = session or requests.Session()

    def _post_json(self, *, url: str, payload: dict[str, Any], headers: dict[str, str]) -> dict[str, Any]:
        last_exc = None
        for attempt in range(1, DEFAULT_RETRY_ATTEMPTS + 1):
            try:
                response = self.session.post(
                    url,
                    json=payload,
                    headers=headers,
                    timeout=DEFAULT_TIMEOUT_SECONDS,
                )
                status_code = getattr(response, "status_code", 200)
                if status_code == 429 and attempt < DEFAULT_RETRY_ATTEMPTS:
                    time.sleep(_retry_delay_seconds(response.headers.get("Retry-After"), attempt))
                    continue
                response.raise_for_status()
                return response.json()
            except requests.exceptions.ConnectionError as exc:
                last_exc = exc
                if not _should_fallback_to_curl(exc):
                    raise
                return _post_with_curl(url=url, payload=payload, headers=headers)
        if last_exc is not None:
            raise last_exc
        raise RuntimeError("Abel CAP request exhausted retries without a response.")

    def discover_parents(self, *, node_id: str, limit: int, api_key: str) -> list[dict[str, Any]]:
        payload = self._post_cap(
            verb="traverse.parents",
            params={"node_id": normalize_public_node_id(node_id), "top_k": min(limit, 20)},
            api_key=api_key,
        )
        return _extract_items(payload)

    def markov_blanket(self, *, node_id: str, limit: int, api_key: str) -> list[dict[str, Any]]:
        payload = self._post_cap(
            verb="graph.markov_blanket",
            params={"node_id": normalize_public_node_id(node_id), "max_neighbors": min(limit, 20)},
            api_key=api_key,
        )
        return _extract_items(payload)

    def fetch_bars(
        self,
        *,
        symbols: list[str],
        start: str | None,
        end: str | None,
        timeframe: str,
        limit: int | None,
        fields: list[str] | None,
        api_key: str,
    ) -> Any:
        payload = self._post_market(
            endpoint="day_bar",
            body={
                "symbols": [
                    normalize_public_node_id(symbol, default_field="price").split(".")[0]
                    for symbol in symbols
                ],
                "start": _serialize_timestamp(start),
                "end": _serialize_timestamp(end),
                "timeframe": timeframe,
                "limit": limit,
                "fields": _normalize_market_fields(fields),
            },
            api_key=api_key,
        )
        items = payload.get("data") or payload.get("result") or []
        if isinstance(items, dict):
            items = items.get("items") or items.get("bars") or []
        return items

    def _post_cap(self, *, verb: str, params: dict[str, Any], api_key: str) -> dict[str, Any]:
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": api_key
            if api_key.lower().startswith("bearer ")
            else f"Bearer {api_key}",
        }
        return self._post_json(
            url=f"{self.cap_base_url}/cap",
            payload={
                "cap_version": CAP_VERSION,
                "request_id": str(uuid.uuid4()),
                "verb": verb,
                "params": params,
                "context": {
                    "graph_ref": {
                        "graph_id": DEFAULT_GRAPH_ID,
                        "graph_version": DEFAULT_GRAPH_VERSION,
                    }
                },
            },
            headers=headers,
        )

    def _post_market(self, *, endpoint: str, body: dict[str, Any], api_key: str) -> dict[str, Any]:
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": api_key
            if api_key.lower().startswith("bearer ")
            else f"Bearer {api_key}",
        }
        return self._post_json(
            url=f"{self.cap_base_url}/market/{endpoint}",
            payload=body,
            headers=headers,
        )


def _should_fallback_to_curl(exc: requests.exceptions.ConnectionError) -> bool:
    if sys.platform != "win32":
        return False
    if shutil.which("curl.exe") is None:
        return False
    message = str(exc).lower()
    return "connectionreseterror" in message or "10054" in message or "connection aborted" in message


def _retry_delay_seconds(retry_after: str | None, attempt: int) -> float:
    if retry_after:
        try:
            return max(1.0, float(retry_after))
        except ValueError:
            pass
    return float(min(2 ** (attempt - 1), 8))


def _post_with_curl(*, url: str, payload: dict[str, Any], headers: dict[str, str]) -> dict[str, Any]:
    curl_path = shutil.which("curl.exe")
    if curl_path is None:
        raise RuntimeError("curl.exe is required for CAP fallback transport on Windows.")

    command = [
        curl_path,
        "--silent",
        "--show-error",
        "--fail-with-body",
        "-X",
        "POST",
        url,
        "--connect-timeout",
        str(DEFAULT_TIMEOUT_SECONDS),
        "--max-time",
        str(DEFAULT_TIMEOUT_SECONDS),
        "--retry",
        "3",
        "--retry-all-errors",
        "--retry-delay",
        "1",
        "--http1.1",
        "--data-binary",
        json.dumps(payload),
    ]
    for key, value in headers.items():
        command.extend(["-H", f"{key}: {value}"])

    result = subprocess.run(command, capture_output=True, text=True, check=False)
    if result.returncode != 0:
        stderr = (result.stderr or result.stdout or "curl fallback failed").strip()
        raise RuntimeError(f"CAP curl fallback failed: {stderr}")
    return json.loads(result.stdout)


def _extract_items(payload: dict[str, Any]) -> list[dict[str, Any]]:
    result = payload.get("result")
    if isinstance(result, dict):
        for key in ("items", "nodes", "neighbors", "markov_blanket"):
            value = result.get(key)
            if isinstance(value, list):
                return [item for item in value if isinstance(item, dict)]
    if isinstance(result, list):
        return [item for item in result if isinstance(item, dict)]
    return []



"""Dashboard HTML generator — reads config + trade logs, renders Jinja2 template.

Three-page layout: Live (default) / Portfolio / Strategy drill-down.
"""
from __future__ import annotations

from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
from jinja2 import Environment, FileSystemLoader

from abel_edge.config import load_config
from abel_edge.dashboard._helpers import fmt_pnl_pct, fmt_dollar
from abel_edge.dashboard.components import (
    compute_metrics,
    yearly_metrics,
    live_metrics,
    equity_chart,
    drawdown_chart,
    rolling_sharpe_chart,
    daily_pnl_chart,
    monthly_heatmap,
)
from abel_edge.dashboard.strategy_data import prepare_strategy, tracked_ticker_item

TEMPLATES_DIR = Path(__file__).parent / "templates"


def _build_env(*, autoescape: bool = False) -> Environment:
    env = Environment(
        loader=FileSystemLoader(str(TEMPLATES_DIR)), autoescape=autoescape,
    )
    env.globals["fmt_pnl_pct"] = fmt_pnl_pct
    env.globals["fmt_dollar"] = fmt_dollar
    return env


def _load_trade_log(path: str) -> pd.DataFrame | None:
    try:
        df = pd.read_csv(path, parse_dates=["date"])
    except (FileNotFoundError, pd.errors.EmptyDataError):
        return None
    if pd.api.types.is_datetime64tz_dtype(df["date"]):
        df["date"] = df["date"].dt.tz_localize(None)
    return df


def _prepare_strategy(s_cfg: dict) -> dict:
    df = _load_trade_log(s_cfg["trade_log"])
    empty = {
        "id": s_cfg["id"], "name": s_cfg["name"],
        "color": s_cfg["color"], "asset": s_cfg["asset"],
        "has_data": False, "metrics": {},
        "health_dots": "", "health_class": "",
    }
    if df is None or len(df) == 0:
        return empty

    pnl = df["pnl"].values.astype(float)
    positions = (df["position"].values.astype(float)
                 if "position" in df.columns else np.zeros(len(pnl)))
    source = (df["source"].values if "source" in df.columns
              else np.array(["backfill"] * len(pnl)))
    asset_close = (df["asset_close"].values.astype(float)
                   if "asset_close" in df.columns else None)
    dates = pd.DatetimeIndex(df["date"])
    cum_pnl = np.cumsum(pnl)
    name = s_cfg["name"]
    color = s_cfg["color"]

    # Validation
    validation = None
    try:
        from abel_edge.validation.gate import validate_strategy
        import tempfile
        with tempfile.NamedTemporaryFile(
            suffix=".csv", delete=False, mode="w"
        ) as f:
            vdf = pd.DataFrame({
                "date": dates, "pnl": pnl, "position": positions,
            })
            vdf.to_csv(f.name, index=False)
            vr = validate_strategy(f.name, positions_col="position")
            Path(f.name).unlink(missing_ok=True)
            validation = {
                "verdict": vr["verdict"],
                "score": vr["score"],
                "failures": vr.get("failures", []),
                "triangle": vr.get("triangle", {}),
                "metrics": {
                    k: round(v, 4) if isinstance(v, float) else v
                    for k, v in vr.get("metrics", {}).items()
                    if k in (
                        "sharpe", "lo_adjusted", "ic", "omega",
                        "max_dd", "dsr", "pbo", "oos_is",
                        "neg_roll_frac", "loss_years", "total_pnl",
                        "calmar", "bootstrap_p", "ic_stability",
                        "sharpe_lo_ratio", "sortino", "ic_hit_rate",
                    )
                },
            }
    except Exception:
        pass

    # Health dots (live vs backtest degradation)
    health_dots, health_class = "", ""
    bt_sharpe = compute_metrics(pnl)["sharpe"]
    live_data = live_metrics(dates, pnl, positions, source, asset_close)
    if live_data and live_data["n_days"] > 10 and bt_sharpe > 0:
        ratio = live_data["sharpe"] / bt_sharpe
        n_dots = max(1, min(5, int(ratio * 5)))
        health_dots = "\u25cf" * n_dots + "\u25cb" * (5 - n_dots)
        health_class = (
            "health-good" if ratio > 0.7
            else "health-warn" if ratio > 0.4
            else "health-bad"
        )
    elif live_data and live_data["n_days"] > 0:
        health_dots = "\u25cb" * 5
        health_class = "health-warn"

    return {
        "id": s_cfg["id"],
        "name": name,
        "color": color,
        "asset": s_cfg["asset"],
        "badge": s_cfg.get("badge", ""),
        "has_data": True,
        "metrics": compute_metrics(pnl),
        "live": live_data,
        "validation": validation,
        "yearly": yearly_metrics(dates, pnl),
        "health_dots": health_dots,
        "health_class": health_class,
        # Charts
        "equity_json": equity_chart(dates, cum_pnl, name, color),
        "drawdown_json": drawdown_chart(dates, cum_pnl, name),
        "rolling_sharpe_json": rolling_sharpe_chart(dates, pnl, name),
        "daily_pnl_json": daily_pnl_chart(dates, pnl, name),
        "monthly_json": monthly_heatmap(dates, pnl, name),
        "latest_date": str(dates[-1].date()),
        "latest_position": float(positions[-1]),
        "latest_pnl": float(pnl[-1]),
        "prev_position": float(positions[-2]) if len(positions) > 1 else 0,
    }


def _sparkline(values: list[float], width: int = 7) -> str:
    """Unicode block sparkline."""
    blocks = " \u2581\u2582\u2583\u2584\u2585\u2586\u2587\u2588"
    if not values or all(v == 0 for v in values):
        return "\u2581" * width
    mn, mx = min(values), max(values)
    rng = mx - mn if mx > mn else 1
    return "".join(
        blocks[min(8, max(0, int((v - mn) / rng * 7) + 1))]
        for v in values[-width:]
    )


def _build_portfolio(strategies: list[dict], settings: dict) -> dict:
    """Aggregate portfolio-level metrics for the Live page."""
    now = datetime.now()

    # Collect latest signals with per-strategy YTD/MTD
    signals_active = []
    signals_flat = []
    today_pnl = 0.0
    mtd_pnl = 0.0
    total_pnl = 0.0
    n_active = 0
    current_month = now.month
    current_year = now.year
    portfolio_id = next(
        (c["id"] for c in _strat_cfgs if c.get("asset") == "MULTI"), None
    )

    for s in strategies:
        if not s["has_data"]:
            continue
        pos = s["latest_position"]
        pnl_today = s["latest_pnl"]
        changed = abs(pos - s["prev_position"]) > 0.01

        strat_ytd = 0.0
        strat_mtd = 0.0
        cfg = next((sc for sc in _strat_cfgs if sc["id"] == s["id"]), None)
        if cfg:
            df_s = _load_trade_log(cfg["trade_log"])
            if df_s is not None:
                df_s["date"] = pd.to_datetime(df_s["date"])
                df_s["_ds"] = df_s["date"].dt.strftime("%Y-%m-%d")
                if "source" in df_s.columns:
                    df_s["_sr"] = df_s["source"].map(
                        {"live": 1, "backfill": 0}
                    ).fillna(0)
                    df_s = df_s.sort_values(["_ds", "_sr"])
                df_s = df_s.drop_duplicates(subset=["_ds"], keep="last")
                ym = df_s["date"].dt.year == current_year
                strat_ytd = float(df_s.loc[ym, "pnl"].sum())
                strat_mtd = float(
                    df_s.loc[ym & (df_s["date"].dt.month == current_month),
                             "pnl"].sum()
                )

        if s["id"] == portfolio_id:
            today_pnl = pnl_today
            total_pnl = strat_ytd
            mtd_pnl = strat_mtd
            continue

        sig = {
            "name": s["name"],
            "color": s["color"],
            "position": pos,
            "today_pnl": pnl_today,
            "ytd_pnl": strat_ytd,
            "mtd_pnl": strat_mtd,
            "changed": changed,
            "id": s["id"],
        }
        if abs(pos) > 0.01:
            signals_active.append(sig)
            n_active += 1
        else:
            signals_flat.append(sig)

    signals_active.sort(key=lambda x: x["position"], reverse=True)

    # Recent days + ledger: both dedup (strategy, date). Extracted to portfolio.py
    # so this file stays under the 400-line structural limit.
    from abel_edge.dashboard.portfolio import build_recent_days
    recent_days, pnl_history = build_recent_days(
        strategies, _strat_cfgs, _load_trade_log,
    )
    if recent_days:
        recent_days[0]["spark"] = _sparkline(list(reversed(pnl_history)))

    # Asset prices (latest from trade logs)
    prices = {}
    seen_assets = set()
    for s in strategies:
        if not s["has_data"] or s["asset"] in seen_assets:
            continue
        df = _load_trade_log(
            next(
                (sc["trade_log"] for sc in _strat_cfgs
                 if sc["id"] == s["id"]),
                "",
            )
        )
        if df is not None and "asset_close" in df.columns:
            price = df["asset_close"].iloc[-1]
            if price > 0:
                prices[s["asset"]] = price
                seen_assets.add(s["asset"])

    # Stale check
    latest_dates = []
    for s in strategies:
        if s["has_data"]:
            latest_dates.append(pd.Timestamp(s["latest_date"]))
    if latest_dates:
        most_recent = max(latest_dates)
        stale_hours = (pd.Timestamp(now) - most_recent).total_seconds() / 3600
        if stale_hours < 1:
            stale_label = "just now"
        elif stale_hours < 24:
            stale_label = f"{int(stale_hours)}h ago"
        else:
            stale_label = f"{int(stale_hours / 24)}d ago"
    else:
        stale_hours = 999
        stale_label = "unknown"

    # Live performance summary per strategy
    live_perf = []
    for s in strategies:
        if not s["has_data"]:
            continue
        cfg_match = next(
            (sc for sc in _strat_cfgs if sc["id"] == s["id"]), None,
        )
        if not cfg_match:
            continue
        df = _load_trade_log(cfg_match["trade_log"])
        if df is None or "source" not in df.columns:
            continue
        live = df[df["source"] == "live"]
        if len(live) < 2:
            continue
        lp = live["pnl"].values.astype(float)
        lpos = live["position"].values.astype(float) if "position" in live.columns else np.zeros(len(lp))
        cum = np.cumsum(lp)
        eq = 1.0 + cum
        peak = np.maximum.accumulate(eq)
        dd = (peak - eq) / peak
        max_dd = float(np.max(dd)) if len(dd) > 0 else 0
        std = np.std(lp, ddof=1) if len(lp) > 1 else 0
        sharpe = float(np.mean(lp) / std * np.sqrt(252)) if std > 0 else 0
        active = lp[np.abs(lp) > 1e-10]
        win_rate = float(np.mean(active > 0)) if len(active) > 0 else 0
        live_perf.append({
            "name": s["name"],
            "color": s["color"],
            "id": s["id"],
            "days": len(lp),
            "sharpe": round(sharpe, 2),
            "pnl": round(float(cum[-1]), 4),
            "max_dd": round(max_dd, 4),
            "win_rate": round(win_rate, 3),
            "active_days": int(np.sum(np.abs(lpos) > 0.01)),
        })
    live_perf.sort(key=lambda x: x["sharpe"], reverse=True)

    from abel_edge.dashboard.portfolio import build_ledger
    since = settings.get("paper_trading_start")  # e.g. "2026-03-01"
    ledger_days = int(settings.get("ledger_days", 30))
    ledger = build_ledger(
        strategies, _strat_cfgs, _load_trade_log,
        since_date=since, n_days=ledger_days,
    )

    signals_flat.sort(key=lambda x: x["ytd_pnl"], reverse=True)
    return {
        "today_pnl": today_pnl,
        "mtd_pnl": mtd_pnl,
        "total_pnl": total_pnl,
        "n_active": n_active,
        "n_flat": len(signals_flat),
        "n_strats": len([s for s in strategies if s["has_data"]]),
        "signals_active": signals_active,
        "signals_flat": signals_flat,
        "recent_days": recent_days,
        "prices": prices,
        "stale_hours": stale_hours,
        "stale_label": stale_label,
        "live_perf": live_perf,
        "ledger": ledger,
    }


# Module-level for _build_portfolio to access
_strat_cfgs: list[dict] = []


def generate(config_path: str, output_path: str, *, bars_loader=None) -> None:
    """Generate the dashboard HTML.

    `bars_loader` is accepted for API compatibility with main-branch callers.
    This dashboard reads asset_return from trade logs directly, so the loader
    isn't used here. Ignored rather than raising.
    """
    del bars_loader  # accepted for CLI compatibility, not used
    global _strat_cfgs
    cfg = load_config(config_path)
    strat_list = cfg["strategies"]
    _strat_cfgs = strat_list

    if len(strat_list) > 1:
        from concurrent.futures import ThreadPoolExecutor
        with ThreadPoolExecutor(
            max_workers=min(len(strat_list), 8)
        ) as pool:
            strategies = list(pool.map(_prepare_strategy, strat_list))
    else:
        strategies = [_prepare_strategy(s) for s in strat_list]

    portfolio = _build_portfolio(strategies, cfg.get("settings", {}))

    env = _build_env(autoescape=False)
    template = env.get_template("base.html")
    html = template.render(
        strategies=strategies,
        portfolio=portfolio,
        settings=cfg.get("settings", {}),
        generated_at=datetime.now().strftime("%Y-%m-%d %H:%M"),
        initial_strategy_id=None,
        initial_section="backtest",
    )
    Path(output_path).write_text(html, encoding="utf-8")


def generate_strategy_dashboard(
    config_path: str | None,
    output_path: str,
    strategy_id: str,
    *,
    bars_loader=None,
    initial_section: str = "paper",
) -> None:
    """Generate the standard dashboard surface focused on one strategy."""
    cfg = load_config(config_path)
    strategies_cfg = [s for s in cfg["strategies"] if s["id"] == strategy_id]
    if not strategies_cfg:
        available = ", ".join(s["id"] for s in cfg["strategies"]) or "none"
        raise ValueError(
            f"Strategy '{strategy_id}' not found in strategies.yaml. Available: {available}"
        )

    strategies = [
        prepare_strategy(strategies_cfg[0], settings=cfg["settings"], bars_loader=bars_loader)
    ]
    env = _build_env(autoescape=False)
    template = env.get_template("tracking.html")
    html = template.render(
        selected_strategy=strategies[0],
        settings=cfg["settings"],
        generated_at=datetime.now().strftime("%Y-%m-%d %H:%M"),
    )
    Path(output_path).write_text(html, encoding="utf-8")


def generate_signal_demo(
    config_path: str | None, output_path: str, strategy_id: str, *, bars_loader=None
) -> None:
    """Generate a single-strategy Signal Demo page."""
    cfg = load_config(config_path)
    strategies_cfg = [s for s in cfg["strategies"] if s["id"] == strategy_id]
    if not strategies_cfg:
        available = ", ".join(s["id"] for s in cfg["strategies"]) or "none"
        raise ValueError(
            f"Strategy '{strategy_id}' not found in strategies.yaml. Available: {available}"
        )

    selected_strategy = prepare_strategy(
        strategies_cfg[0], settings=cfg["settings"], bars_loader=bars_loader
    )
    tracked_tickers = [
        tracked_ticker_item(s, settings=cfg["settings"], bars_loader=bars_loader)
        for s in cfg["strategies"]
    ]
    env = _build_env(autoescape=False)
    template = env.get_template("signal_demo.html")
    html = template.render(
        selected_strategy=selected_strategy,
        tracked_tickers=tracked_tickers,
        settings=cfg["settings"],
        generated_at=datetime.now().strftime("%Y-%m-%d %H:%M"),
    )
    Path(output_path).write_text(html, encoding="utf-8")


def generate_tracking_page(
    config_path: str | None, output_path: str, strategy_id: str, *, bars_loader=None
) -> None:
    """Generate a single-strategy paper-tracking page."""
    generate_strategy_dashboard(
        config_path,
        output_path,
        strategy_id,
        bars_loader=bars_loader,
        initial_section="paper",
    )



"""Minimal DecisionContext SMA crossover example."""

from __future__ import annotations

from abel_edge.engine.base import StrategyEngine


class SMAEngine(StrategyEngine):
    """Simple moving-average crossover on the primary target feed."""

    def __init__(self, context: dict | None = None) -> None:
        super().__init__(context=context)
        self.fast = 10
        self.slow = 30

    def compute_decisions(self, ctx):
        close = ctx.target.series("close")
        fast_ma = close.rolling(self.fast, min_periods=self.fast).mean()
        slow_ma = close.rolling(self.slow, min_periods=self.slow).mean()
        next_position = (fast_ma > slow_ma).astype(float).fillna(0.0)
        if len(next_position) > 0:
            next_position.iloc[: self.slow] = 0.0
        return ctx.decisions(next_position)



"""Abel price data helpers."""

from __future__ import annotations

import pandas as pd

from abel_edge.plugins.abel.client import AbelClient
from abel_edge.plugins.abel.credentials import MissingAbelApiKeyError, require_api_key


def fetch_bars(
    *,
    symbols: list[str],
    start=None,
    end=None,
    timeframe: str = "1d",
    limit: int | None = None,
    fields: list[str] | None = None,
    config: dict | None = None,
    client: AbelClient | None = None,
) -> pd.DataFrame:
    env_path = (config or {}).get("env_path", ".env")
    try:
        api_key = require_api_key(env_path=env_path)
    except MissingAbelApiKeyError as e:
        raise MissingAbelApiKeyError(
            f"{e} Or set price_data.adapter to 'csv' for local bar data."
        ) from e
    abel = client or AbelClient(env_path=env_path)
    payload = abel.fetch_bars(
        symbols=symbols,
        start=start,
        end=end,
        timeframe=timeframe,
        limit=limit,
        fields=fields,
        api_key=api_key,
    )
    return pd.DataFrame(payload)



from pathlib import Path
import importlib
import sys

import pandas as pd
from click.testing import CliRunner

from abel_edge.config import load_config
from abel_edge.engine.ledger import read_trade_log, write_trade_log
from abel_edge.engine.trader import paper_run_one


DECISION_ENGINE_CODE = """
from __future__ import annotations

import pandas as pd

from abel_edge.engine.base import StrategyEngine


class CountingDecisionEngine(StrategyEngine):
    calls = 0

    def compute_decisions(self, ctx):
        type(self).calls += 1
        close = ctx.target.series("close")
        next_position = pd.Series(
            [idx * 0.25 for idx in range(len(close))],
            index=close.index,
        )
        return ctx.decisions(next_position)
""".strip()


def _clear_strategy_modules() -> None:
    for name in list(sys.modules):
        if name == "strategies" or name.startswith("strategies."):
            sys.modules.pop(name, None)
    importlib.invalidate_caches()


def _write_decision_project(root: Path) -> None:
    _clear_strategy_modules()
    strategy_dir = root / "strategies" / "decision_once"
    strategy_dir.mkdir(parents=True)
    (root / "strategies" / "__init__.py").write_text("", encoding="utf-8")
    (strategy_dir / "__init__.py").write_text("", encoding="utf-8")
    (strategy_dir / "engine.py").write_text(DECISION_ENGINE_CODE, encoding="utf-8")
    (root / "data").mkdir()
    (root / "data" / "ethusd.csv").write_text(
        "timestamp,close\n"
        "2026-01-01T00:00:00Z,100\n"
        "2026-01-02T00:00:00Z,110\n"
        "2026-01-03T00:00:00Z,120\n"
        "2026-01-04T00:00:00Z,130\n",
        encoding="utf-8",
    )
    (root / "strategies.yaml").write_text(
        """
settings:
  price_data:
    default_source: csv
    default_timeframe: 1d
strategies:
  - id: decision_once
    name: "Decision Once"
    asset: ETHUSD
    color: "#2563EB"
    engine: strategies.decision_once.engine
    trade_log: data/trade_log_decision_once.csv
    paper_log: data/paper_log_decision_once.csv
    price_data:
      source: csv
      path: data/ethusd.csv
""".strip()
        + "\n",
        encoding="utf-8",
    )


def test_paper_run_one_default_signal_computes_once_and_rolls_position(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        root = Path.cwd()
        _write_decision_project(root)
        sys.path.insert(0, str(root))

        try:
            dates = pd.to_datetime(
                ["2026-01-01T00:00:00Z", "2026-01-02T00:00:00Z"],
                utc=True,
            )
            write_trade_log(
                dates,
                [0.0, 0.10],
                [0.0, 0.0],
                [0.0, 0.0],
                "data/trade_log_decision_once.csv",
                close_prices=[100.0, 110.0],
                next_positions=[0.0, 0.25],
            )
            cfg = load_config()

            result = paper_run_one(
                cfg["strategies"][0],
                settings=cfg.get("settings"),
                as_of="2026-01-04T00:00:00Z",
            )

            engine_module = importlib.import_module("strategies.decision_once.engine")
            assert engine_module.CountingDecisionEngine.calls == 1
            assert result["n_rows"] == 2
            paper_df = read_trade_log("data/paper_log_decision_once.csv")
            assert list(paper_df["date"].dt.strftime("%Y-%m-%d")) == [
                "2026-01-03",
                "2026-01-04",
            ]
            assert list(paper_df["position"].round(2)) == [0.25, 0.50]
            assert list(paper_df["next_position"].round(2)) == [0.50, 0.75]
        finally:
            sys.path.pop(0)



"""Trade log read/write. Single source of truth for trade log CSV format."""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd


REQUIRED_COLUMNS = ("date", "pnl", "position", "cum_return", "source")


def read_trade_log(path: str | Path) -> pd.DataFrame:
    """Read a trade log CSV. Returns DataFrame with standard columns.

    Date parsing uses `format="mixed"` because backfill rows write midnight
    timestamps ("2026-04-16 00:00:00+00:00") while live rows carry
    sub-second ISO-8601 ("2026-04-17T05:55:06.150276+00:00"). The default
    strptime fallback on mixed formats raises, which blocked every
    subsequent `abel-edge run` once any live row existed.
    """
    df = pd.read_csv(path)
    if "date" in df.columns:
        df["date"] = pd.to_datetime(df["date"], utc=True, format="mixed")
    if "decision_time" in df.columns:
        df["decision_time"] = pd.to_datetime(df["decision_time"], utc=True, format="mixed")
    if "effective_time" in df.columns:
        df["effective_time"] = pd.to_datetime(df["effective_time"], utc=True, format="mixed")
    return df


def write_trade_log(
    dates: pd.DatetimeIndex,
    asset_returns: np.ndarray,
    pnl: np.ndarray,
    positions: np.ndarray,
    path: str | Path,
    source: str = "backfill",
    decision_times: pd.DatetimeIndex | None = None,
    effective_times: pd.DatetimeIndex | None = None,
    close_prices: np.ndarray | None = None,
    next_positions: np.ndarray | None = None,
    gross_pnl: np.ndarray | None = None,
    turnover: np.ndarray | None = None,
    execution_cost: np.ndarray | None = None,
) -> None:
    """Write a trade log CSV from strategy output arrays.

    Args:
        dates: Trading dates
        asset_returns: Daily simple returns of the underlying asset
        pnl: Daily net PnL after execution costs
        positions: Daily position sizes
        path: Output CSV path
        source: "backfill" or "live"
    """
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)

    df = pd.DataFrame(
        {
            "date": dates,
            "asset_return": asset_returns,
            "pnl": pnl,
            "position": positions,
            "source": source,
        }
    )
    if decision_times is not None:
        df["decision_time"] = decision_times
    if effective_times is not None:
        df["effective_time"] = effective_times
    if close_prices is not None:
        df["close"] = close_prices
    if next_positions is not None:
        df["next_position"] = next_positions
    if gross_pnl is not None:
        df["gross_pnl"] = gross_pnl
    if turnover is not None:
        df["turnover"] = turnover
    if execution_cost is not None:
        df["execution_cost"] = execution_cost
    if path.exists():
        existing = read_trade_log(path)
        if "source" in existing.columns:
            existing["source"] = existing["source"].fillna("backfill").astype(str)
            existing["date"] = pd.to_datetime(existing["date"], utc=True)
            live_rows = existing[existing["source"].str.lower() == "live"].copy()
            if not live_rows.empty:
                live_rows = _dedupe_trade_rows(live_rows)
                df["date"] = pd.to_datetime(df["date"], utc=True)
                df = pd.concat([df, live_rows], ignore_index=True, sort=False)

    df = _dedupe_trade_rows(df)
    df["cum_return"] = df["pnl"].to_numpy(dtype=float).cumsum()
    df.to_csv(path, index=False)


def append_trade_log_rows(path: str | Path, rows: list[dict]) -> pd.DataFrame:
    """Append live paper-trading rows and recompute cumulative return."""
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)

    incoming = pd.DataFrame(rows)
    if incoming.empty:
        return read_trade_log(path) if path.exists() else incoming

    incoming["date"] = pd.to_datetime(incoming["date"], utc=True)

    if path.exists():
        existing = read_trade_log(path)
    else:
        existing = pd.DataFrame(columns=incoming.columns)

    combined = pd.concat([existing, incoming], ignore_index=True, sort=False)
    combined = _dedupe_trade_rows(combined)
    combined["pnl"] = combined["pnl"].astype(float)
    combined["cum_return"] = combined["pnl"].to_numpy(dtype=float).cumsum()
    combined.to_csv(path, index=False)
    return combined.reset_index(drop=True)


def _dedupe_trade_rows(df: pd.DataFrame) -> pd.DataFrame:
    combined = df.copy()
    if "date" in combined.columns:
        combined["date"] = pd.to_datetime(combined["date"], utc=True)
    if "source" not in combined.columns:
        combined["source"] = "backfill"
    combined["source"] = combined["source"].fillna("backfill").astype(str)
    combined["_src_rank"] = combined["source"].str.lower().map({"live": 1, "backfill": 0}).fillna(0)
    combined = combined.sort_values(["date", "_src_rank"], kind="mergesort")
    combined = combined.drop_duplicates(subset=["date"], keep="last")
    return combined.drop(columns="_src_rank").reset_index(drop=True)



"""Abel discovery helpers used by the CLI."""

from __future__ import annotations

from datetime import datetime, timezone
from typing import Any

from abel_edge.plugins.abel.client import AbelClient, split_public_node_id
from abel_edge.plugins.abel.credentials import MissingAbelApiKeyError, require_api_key


def discover_graph_nodes(
    node_id: str,
    *,
    mode: str = "parents",
    limit: int = 10,
    env_path: str = ".env",
    client: AbelClient | None = None,
) -> str:
    payload = discover_graph_payload(
        node_id,
        mode=mode,
        limit=limit,
        env_path=env_path,
        client=client,
    )
    return render_discovery_payload(payload, mode=mode)


def discover_graph_payload(
    node_id: str,
    *,
    mode: str = "all",
    limit: int = 10,
    env_path: str = ".env",
    client: AbelClient | None = None,
) -> dict[str, Any]:
    try:
        api_key = require_api_key(env_path=env_path)
    except MissingAbelApiKeyError as e:
        raise MissingAbelApiKeyError(
            f"{e} Optionally set ABEL_CAP_BASE_URL to target a non-default CAP endpoint."
        ) from e
    abel = client or AbelClient(env_path=env_path)
    limit = min(max(limit, 1), 20)

    if mode == "all":
        parents = _discover_mode_items(
            node_id=node_id,
            mode="parents",
            limit=limit,
            api_key=api_key,
            client=abel,
        )
        blanket_items = _discover_mode_items(
            node_id=node_id,
            mode="mb",
            limit=limit,
            api_key=api_key,
            client=abel,
        )
        return _build_discovery_payload(
            node_id,
            parents=parents,
            blanket_items=blanket_items,
        )

    items = _discover_mode_items(
        node_id=node_id,
        mode=mode,
        limit=limit,
        api_key=api_key,
        client=abel,
    )
    target_asset, target_field = split_public_node_id(node_id)
    target_node = f"{target_asset}.{target_field}"
    if mode == "parents":
        return {
            "ticker": target_asset,
            "target_asset": target_asset,
            "target_node": target_node,
            "source": "abel_live",
            "mode": mode,
            "parents": items,
            "blanket_new": [],
            "children": [],
            "K_discovery": len(items),
            "created_at": _now(),
        }
    if mode == "mb":
        payload = _build_discovery_payload(node_id, parents=[], blanket_items=items)
        payload["mode"] = mode
        payload["K_discovery"] = 0
        return payload
    raise ValueError(f"Unsupported mode '{mode}'.")


def render_discovery_payload(payload: dict[str, Any], *, mode: str = "all") -> str:
    if mode == "parents":
        return _render_parents(payload.get("parents", []))
    if mode == "mb":
        return _render_markov_blanket(
            payload.get("blanket_new", []),
            payload.get("children", []),
        )
    if mode == "all":
        return _render_combined(payload)
    raise ValueError(f"Unsupported mode '{mode}'.")


def _render_parents(items: list[dict[str, Any]]) -> str:
    lines = ["parents:"]
    for item in items[:20]:
        ticker = str(item.get("ticker", "")).strip()
        field = str(item.get("field", "")).strip()
        if not ticker or not field:
            continue
        lines.append(f"  - ticker: {ticker}")
        lines.append(f"    field: {field}")
    return "\n".join(lines)


def _render_markov_blanket(
    blanket_new: list[dict[str, Any]],
    children: list[dict[str, Any]],
) -> str:
    lines = ["markov_blanket:"]
    rendered_items = []
    for item in children:
        rendered_items.append(
            {
                "ticker": item.get("ticker", ""),
                "field": item.get("field", ""),
                "roles": ["child"],
            }
        )
    rendered_items.extend(blanket_new)
    for item in rendered_items[:20]:
        ticker = str(item.get("ticker", "")).strip()
        field = str(item.get("field", "")).strip()
        if not ticker or not field:
            continue
        roles = [str(role).strip() for role in item.get("roles", []) if str(role).strip()]
        lines.append(f"  - ticker: {ticker}")
        lines.append(f"    field: {field}")
        lines.append(f"    roles: [{', '.join(roles)}]")
    return "\n".join(lines)


def _render_combined(payload: dict[str, Any]) -> str:
    parts = [
        f"ticker: {payload.get('ticker', '')}",
        f"target_node: {payload.get('target_node', '')}",
        f"source: {payload.get('source', '')}",
        f"K_discovery: {payload.get('K_discovery', 0)}",
        _render_parents(payload.get("parents", [])),
        _render_markov_blanket(
            payload.get("blanket_new", []),
            payload.get("children", []),
        ),
    ]
    return "\n\n".join(part for part in parts if part)


def _discover_mode_items(
    *,
    node_id: str,
    mode: str,
    limit: int,
    api_key: str,
    client: AbelClient,
) -> list[dict[str, Any]]:
    if mode == "parents":
        raw_items = client.discover_parents(node_id=node_id, limit=limit, api_key=api_key)
        return _normalize_items(raw_items)
    if mode == "mb":
        raw_items = client.markov_blanket(node_id=node_id, limit=limit, api_key=api_key)
        return _normalize_items(raw_items)
    raise ValueError(f"Unsupported mode '{mode}'.")


def _build_discovery_payload(
    node_id: str,
    *,
    parents: list[dict[str, Any]],
    blanket_items: list[dict[str, Any]],
) -> dict[str, Any]:
    target_asset, target_field = split_public_node_id(node_id)
    target_node = f"{target_asset}.{target_field}"
    parent_keys = {item["node_id"] for item in parents}
    children: list[dict[str, Any]] = []
    blanket_new: list[dict[str, Any]] = []
    seen_children: set[str] = set()
    seen_blanket: set[str] = set()

    for item in blanket_items:
        key = item["node_id"]
        roles = [str(role).strip() for role in item.get("roles", []) if str(role).strip()]
        if "child" in roles and key not in seen_children:
            children.append(
                {
                    "node_id": item["node_id"],
                    "ticker": item["ticker"],
                    "field": item["field"],
                }
            )
            seen_children.add(key)
            continue
        if key in parent_keys or key in seen_blanket:
            continue
        blanket_new.append(
            {
                "node_id": item["node_id"],
                "ticker": item["ticker"],
                "field": item["field"],
                "roles": roles or ["neighbor"],
            }
        )
        seen_blanket.add(key)

    return {
        "ticker": target_asset,
        "target_asset": target_asset,
        "target_node": target_node,
        "source": "abel_live",
        "parents": parents,
        "blanket_new": blanket_new,
        "children": children,
        "K_discovery": len(parents),
        "created_at": _now(),
    }


def _normalize_items(items: list[dict[str, Any]]) -> list[dict[str, Any]]:
    normalized: list[dict[str, Any]] = []
    for item in items[:20]:
        node_id = _pick_node_id(item)
        if not node_id:
            continue
        ticker, field = split_public_node_id(node_id)
        normalized.append(
            {
                "node_id": f"{ticker}.{field}",
                "ticker": ticker,
                "field": field,
                "roles": _pick_roles(item),
            }
        )
    return normalized


def _pick_node_id(item: dict[str, Any]) -> str:
    for key in ("node_id", "id", "name"):
        value = item.get(key)
        if isinstance(value, str) and value.strip():
            return value.strip()
    return ""


def _pick_roles(item: dict[str, Any]) -> list[str]:
    for key in ("roles", "role", "relationship", "type"):
        value = item.get(key)
        if isinstance(value, list):
            return [str(role).strip() for role in value if str(role).strip()]
        if isinstance(value, str) and value.strip():
            return [value.strip()]
    return ["neighbor"]


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()
